import matplotlib.pyplot as plt

from pysmad.bodies.artificial import Spacecraft
from pysmad.coordinates.states import GCRF, HCW, StateConvert
from pysmad.math.linalg import Vector3D, Vector6D
from pysmad.propagators.relative import Hill
from pysmad.time import Epoch

# Create initial scenario epoch
start_epoch: Epoch = Epoch.from_gregorian(2022, 12, 20, 0, 0, 0)

# Create a desired relative state for the chase vehicle
rel_chase_state = HCW.from_state_vector(Vector6D(-11, 0, 0, 0, 0.0016, 0))

# Create an ECI state for the target to act as the origin
target_state: GCRF = GCRF(start_epoch, Vector3D(42164, 0, 0), Vector3D(0, 3.075, 0))

# Create an ECI state for the chase vehicle using the target ECI state as the origin
chase_state: GCRF = StateConvert.hcw.to_gcrf(rel_chase_state, target_state)

# Create a propagation end epoch after 5 days
end_epoch = start_epoch.plus_days(5)

# Load states for chase and target
chase: Spacecraft = Spacecraft(chase_state)
target: Spacecraft = Spacecraft(target_state)

# Create a Hill propagator to compare RK4 results
rel = Hill(rel_chase_state, target.sma())

hill_r = []
hill_i = []
rk_r = []
rk_i = []

# Propagate
while chase.current_epoch().value < end_epoch.value:

    # Step vehicles and propagator
    chase.step()
    target.step_to_epoch(chase.current_epoch())
    rel.step_by_seconds(chase.propagator.step_size)

    # Calculate the hill state of chase and target after RK4 propagation
    ric = target.hill_position(chase)

    # Store data
    rk_r.append(ric.x)
    rk_i.append(ric.y)
    hill_r.append(rel.state.position.x)
    hill_i.append(rel.state.position.y)

# Overlay RI points of RK4 and Hill
plt.plot(hill_i, hill_r)
plt.plot(rk_i, rk_r)
plt.show()



import unittest
from math import radians

from pysmad.bodies.artificial import Spacecraft
from pysmad.bodies.celestial import Earth
from pysmad.bodies.terrestrial import GroundSite
from pysmad.coordinates.elements import ClassicalElements
from pysmad.coordinates.positions import SphericalPosition
from pysmad.coordinates.states import GCRF, IJK
from pysmad.estimation.obs import GroundObservation
from pysmad.math.constants import SECONDS_IN_DAY
from pysmad.time import Epoch


class TestGroundSite(unittest.TestCase):

    spherical: SphericalPosition = SphericalPosition(Earth.RADIUS, radians(11), radians(48))
    SITE: GroundSite = GroundSite.from_itrf_position(spherical.to_cartesian())
    COES: ClassicalElements = ClassicalElements(Epoch(50449), Earth.RADIUS + 960, 0, radians(97), radians(130.7), 0, 0)
    IJK_STATE: IJK = COES.to_ijk()
    SAT: Spacecraft = Spacecraft(GCRF(IJK_STATE.epoch, IJK_STATE.position, IJK_STATE.velocity))

    def test_angles_and_range(self):
        self.SAT.step_to_epoch(self.SAT.current_epoch().plus_days(600 / SECONDS_IN_DAY))
        az_el_range: GroundObservation = self.SITE.angles_and_range(self.SAT)
        self.assertAlmostEqual(az_el_range.azimuth, 2.4905857059239622)
        self.assertAlmostEqual(az_el_range.elevation, 0.3057319222804495)
        self.assertAlmostEqual(az_el_range.range, 2174.186500131194)



import matplotlib.pyplot as plt

from pysmad.bodies.artificial import Spacecraft
from pysmad.coordinates.states import GCRF
from pysmad.math.linalg import Vector3D
from pysmad.time import Epoch

# Create initial scenario epoch
start_epoch: Epoch = Epoch.from_gregorian(2022, 12, 20, 0, 0, 0)

# Create an ECI state for the spacecraft
passive_state: GCRF = GCRF(start_epoch, Vector3D(42164, 0, 0), Vector3D(0, 3.075, 0))

# Create a propagation end epoch after 2 days
end_epoch = start_epoch.plus_days(2)

# Load states
freeflight: Spacecraft = Spacecraft(passive_state)
maneuver: Spacecraft = Spacecraft(passive_state)

rk_r = []
rk_i = []
maneuver_complete: bool = False
maneuver_direction: Vector3D = Vector3D(0, 0.003, 0)

# Propagate
while freeflight.current_epoch().value < end_epoch.value:

    # Step vehicles and propagator
    freeflight.step()
    maneuver.step_to_epoch(freeflight.current_epoch())

    # Calculate the hill state of chase and target after RK4 propagation
    ric = freeflight.hill_position(maneuver)

    # Perform a maneuver after one day
    if freeflight.current_epoch().value > start_epoch.value + 1 and not maneuver_complete:
        maneuver.finite_maneuver(maneuver_direction)
        freeflight.step_to_epoch(maneuver.current_epoch())
        maneuver_complete = True

    # Store data
    rk_r.append(ric.x)
    rk_i.append(ric.y)

# Overlay RI points of RK4 and Hill
plt.plot(rk_i, rk_r)
plt.show()



#: number of seconds in one solar day in :math:`\frac{sec}{day}`
SECONDS_IN_DAY: float = 86400

#: number of hours in one solar day in :math:`\frac{hr}{day}`
HOURS_IN_DAY: float = 24

#: number of minutes equal to one solar day in :math:`\frac{min}{day}`
MINUTES_IN_DAY: float = 1440

#: number of minutes equal to one hour in :math:`\frac{min}{hr}`
MINUTES_IN_HOUR: float = 60

#: number of seconds equal to one hour in :math:`\frac{sec}{hr}`
SECONDS_IN_HOUR: float = 3600

#: number of seconds equal to one minute in :math:`\frac{sec}{min}`
SECONDS_IN_MINUTE: float = 60

#: number of degrees equal to one hour in :math:`\frac{\deg}{hr}`
DEGREES_IN_HOUR: float = 15

#: gravity at sea level in :math:`\frac{km}{sec^2}`
SEA_LEVEL_G: float = 0.00981

#: number of base units in a kilo in :math:`\frac{(base)}{kilo(base)}`
BASE_IN_KILO: float = 1000

#: number of seconds for Earth to complete one rotation in :math:`sec` (found on page 31 of Vallado 4th Edition)
SECONDS_IN_SIDEREAL_DAY: float = 86164.090518



from math import cos, sin, sqrt

from pysmad.bodies.celestial import Earth
from pysmad.coordinates.states import IJK
from pysmad.math.functions import EquationsOfMotion
from pysmad.math.linalg import Vector3D
from pysmad.time import Epoch


class ClassicalElements:
    def __init__(self, epoch: Epoch, a: float, e: float, i: float, raan: float, arg_per: float, ma: float) -> None:
        """used to perform calculations with the classical orbital elements

        :param epoch: epoch for which the elements are valid
        :type epoch: Epoch
        :param a: semi-major axis in km
        :type a: float
        :param e: eccentricity
        :type e: float
        :param i: inclination in radians
        :type i: float
        :param raan: right ascension of the ascending node in radians
        :type raan: float
        :param arg_per: argument of perigee in radians
        :type arg_per: float
        :param ma: mean anomaly in radians
        :type ma: float
        """
        #: epoch for which the element set is valid
        self.epoch: Epoch = epoch.copy()

        #: semi-major axis in km
        self.semimajor_axis: float = a

        #: eccentricity
        self.eccentricity: float = e

        #: inclination in radians
        self.inclination: float = i

        #: right ascension of the ascending node in radians
        self.raan: float = raan

        #: argument of perigee in radians
        self.argument_of_perigee: float = arg_per

        #: mean anomaly in radians
        self.mean_anomaly: float = ma

    @classmethod
    def from_ijk(cls, state: IJK) -> "ClassicalElements":
        r: float = state.position.magnitude()
        v: float = state.velocity.magnitude()
        rdv: float = state.position.dot(state.velocity)

        h: Vector3D = EquationsOfMotion.H.from_r_v(state.position, state.velocity)
        w: Vector3D = h.normalized()
        i: float = EquationsOfMotion.I.from_w(w)
        raan: float = EquationsOfMotion.RAAN.from_w(w)
        a: float = EquationsOfMotion.A.from_mu_r_v(Earth.MU, r, v)
        p: float = EquationsOfMotion.P.from_mu_h(Earth.MU, h.magnitude())
        e: float = EquationsOfMotion.E.from_a_p(a, p)
        n: float = EquationsOfMotion.N.from_a_mu(a, Earth.MU)
        ea: float = EquationsOfMotion.EA.from_rdv_r_a_n(rdv, r, a, n)
        ma: float = EquationsOfMotion.MA.from_ea_e(ea, e)
        u: float = EquationsOfMotion.U.from_r_w(state.position, w)
        nu: float = EquationsOfMotion.NU.from_e_ea(e, ea)
        aop: float = EquationsOfMotion.W.from_u_nu(u, nu)
        return cls(state.epoch, a, e, i, raan, aop, ma)

    def eccentric_anomaly(self) -> float:
        """calculate the eccentric anomaly

        :return: eccentric anomaly in radians
        :rtype: float
        """
        return EquationsOfMotion.EA.from_ma_e(self.mean_anomaly, self.eccentricity)

    def perigee_vector(self) -> Vector3D:
        """calculate the vector pointing from the focus to the lowest point in the orbit

        :return: vector from origin to perigee
        :rtype: Vector3D
        """
        cw: float = cos(self.argument_of_perigee)
        c0: float = cos(self.raan)
        sw: float = sin(self.argument_of_perigee)
        s0: float = sin(self.raan)
        ci: float = cos(self.inclination)
        return Vector3D(cw * c0 - sw * ci * s0, cw * s0 + sw * ci * c0, sw * sin(self.inclination)).normalized()

    def semilatus_rectum_vector(self) -> Vector3D:
        """calculate the vector from the focus to the point 90 degrees off of the perigee vector

        :param epoch: _description_
        :type epoch: Epoch
        :return: semi-latus rectum vector
        :rtype: Vector3D
        """
        cw: float = cos(self.argument_of_perigee)
        c0: float = cos(self.raan)
        sw: float = sin(self.argument_of_perigee)
        s0: float = sin(self.raan)
        ci: float = cos(self.inclination)
        return Vector3D(-sw * c0 - cw * ci * s0, -sw * s0 + cw * ci * c0, cw * sin(self.inclination))

    def to_ijk(self) -> IJK:
        """calculates the cartesian representation of the element set

        :return: inertial state of the element set
        :rtype: IJK
        """
        p: Vector3D = self.perigee_vector()
        q: Vector3D = self.semilatus_rectum_vector()

        ea: float = self.eccentric_anomaly()
        cea: float = cos(ea)
        sea: float = sin(ea)

        r: float = self.semimajor_axis * (1 - self.eccentricity * cea)

        ea_dot: float = (1 / r) * sqrt(Earth.MU / self.semimajor_axis)

        b: float = self.semimajor_axis * sqrt(1 - self.eccentricity * self.eccentricity)

        x_bar: float = self.semimajor_axis * (cea - self.eccentricity)
        y_bar: float = b * sea
        x_bar_dot: float = -self.semimajor_axis * ea_dot * sea
        y_bar_dot: float = b * ea_dot * cea

        return IJK(self.epoch, p.scaled(x_bar).plus(q.scaled(y_bar)), p.scaled(x_bar_dot).plus(q.scaled(y_bar_dot)))



import unittest

from pysmad.coordinates.elements import ClassicalElements
from pysmad.coordinates.states import IJK
from pysmad.math.linalg import Vector3D
from pysmad.time import Epoch


class TestClassicalElements(unittest.TestCase):
    EPOCH: Epoch = Epoch(0)
    POSITION: Vector3D = Vector3D(10000, 40000, -5000)
    VELOCITY: Vector3D = Vector3D(-1.5, 1, -0.1)
    GEO_SMA: float = 42164
    GEO_PERIOD: float = 86163.57055057827
    IJK_STATE: IJK = IJK(EPOCH, POSITION, VELOCITY)

    def test_from_r_v(self):
        coes: ClassicalElements = ClassicalElements.from_ijk(self.IJK_STATE)
        self.assertAlmostEqual(coes.semimajor_axis, 25015.18101846454)
        self.assertAlmostEqual(coes.eccentricity, 0.707977170873199)
        self.assertAlmostEqual(coes.inclination, 0.12166217595729033)
        self.assertAlmostEqual(coes.raan, 3.024483909022929)
        self.assertAlmostEqual(
            coes.argument_of_perigee,
            1.597899323919623,
        )
        self.assertAlmostEqual(coes.mean_anomaly, 2.5172009599614285)



from math import radians

from pysmad.coordinates.positions import ENZ, LLA, PositionConvert, SphericalPosition
from pysmad.coordinates.states import GCRF, ITRF
from pysmad.math.linalg import Vector3D
from pysmad.time import Epoch


class Observation:
    """super class used for space-based and ground-based observations"""

    def __init__(self) -> None:
        pass

    def ijk_position(self) -> Vector3D:
        raise NotImplementedError

    def epoch(self) -> Epoch:
        raise NotImplementedError

    def observer_ijk(self) -> Vector3D:
        raise NotImplementedError


class SpaceObservation(Observation):
    def __init__(self, observer_state: GCRF, observed_direction: Vector3D, r_error: float, ang_error: float) -> None:
        """object used for state estimation when the observer is a space asset

        :param observer_state: inertial state of the observer at the time of the observation
        :type observer_state: GCRF
        :param observed_direction: GCRF direction of the object from the observer
        :type observed_direction: Vector3D
        :param r_error: one-sigma error of the observed range in km
        :type r_error: float
        :param ang_error: one-sigma error of the angles in radians
        :type ang_error: float
        """
        #: inertial state of the observer at the time of the observation
        self.observer_state: GCRF = observer_state.copy()

        #: vector from observer to target in the GCRF frame
        self.observed_direction: Vector3D = observed_direction.copy()

        spherical: SphericalPosition = SphericalPosition.from_cartesian(observed_direction)

        #: magnitude of the observation in km
        self.range: float = spherical.radius

        #: right ascension of the observation in radians
        self.right_ascension: float = spherical.right_ascension

        #: declination of the observation in radians
        self.declination: float = spherical.declination

        #: one-sigma range error of the observation in km
        self.range_error: float = r_error

        #: one-sigma angular error of the observation in radians
        self.angular_error: float = ang_error

    def epoch(self) -> Epoch:
        """get the time the observation was taken

        :return: valid epoch of the observation
        :rtype: Epoch
        """
        return self.observer_state.epoch

    def ijk_position(self) -> Vector3D:
        """calculate the inertial position of the observation

        :return: inertial position of the observation in the IJK frame
        :rtype: Vector3D
        """
        return PositionConvert.gcrf.to_ijk(self.observer_state.position.plus(self.observed_direction), self.epoch())


class GroundObservation(Observation):
    def __init__(self, observer_state: ITRF, observed_direction: Vector3D, r_error: float, ang_error: float) -> None:
        """used to perform operations related to ground site measurements

        :param observer_state: geocentric coordinates of the observer
        :type observer_state: ITRF
        :param observed_direction: ENZ direction of object from observer
        :type observed_direction: Vector3D
        :param r_error: one-sigma error of the observed range in km
        :type r_error: float
        :param ang_error: one-sigma error of the angles in radians
        :type ang_error: float
        """
        #: geocentric state of the observer at the time of the observation
        self.observer_state: ITRF = observer_state.copy()

        #: vector from observer to target in the ENZ frame
        self.observed_direction: Vector3D = observed_direction.copy()

        spherical: SphericalPosition = SphericalPosition.from_cartesian(
            Vector3D(observed_direction.y, observed_direction.x, observed_direction.z)
        )

        #: magnitude of the observation in km
        self.range: float = spherical.radius

        #: azimuth of the observation in radians
        self.azimuth: float = spherical.right_ascension

        #: elevation of the observation in radians
        self.elevation: float = spherical.declination

        #: one-sigma range error of the observation in km
        self.range_error: float = r_error

        #: one-sigma angular error of the observation in radians
        self.angular_error: float = ang_error

    @classmethod
    def from_angles_and_range(
        cls, observer_state: ITRF, az: float, el: float, r: float, r_error: float, ang_error: float
    ) -> "GroundObservation":
        """create an observation from azimuth, elevation, and range

        :param observer_state: geocentric coordinates of the observer
        :type observer_state: ITRF
        :param az: azimuth of the observation in radians
        :type az: float
        :param el: elevation of the observation in radians
        :type el: float
        :param r: magnitude of the observation in km
        :type r: float
        :param r_error: one-sigma range error of the observation in km
        :type r_error: float
        :param ang_error: one-sigma angular error of the observation in radians
        :type ang_error: float
        :return: observation from a terrestrial site
        :rtype: GroundObservation
        """
        nez: Vector3D = SphericalPosition(r, az, el).to_cartesian()
        enz: Vector3D = Vector3D(nez.y, nez.x, nez.z)
        return cls(observer_state, enz, r_error, ang_error)

    def epoch(self) -> Epoch:
        """get the time the observation was taken

        :return: valid epoch of the observation
        :rtype: Epoch
        """
        return self.observer_state.epoch

    def ijk_position(self) -> Vector3D:
        """calculate the inertial position of the observation

        :return: inertial position of the observation in the IJK frame
        :rtype: Vector3D
        """
        lla_site: LLA = PositionConvert.itrf.to_lla(self.observer_state.position)
        itrf_ob: Vector3D = PositionConvert.enz.to_itrf(lla_site, self.observed_direction).plus(
            self.observer_state.position
        )
        return PositionConvert.itrf.to_ijk(itrf_ob, self.epoch())

    def gcrf_position(self) -> Vector3D:
        """calculate the inertial position of the observation

        :return: inertial position of the observation in the GCRF frame
        :rtype: Vector3D
        """
        lla_site: LLA = PositionConvert.itrf.to_lla(self.observer_state.position)
        itrf_ob: Vector3D = PositionConvert.enz.to_itrf(lla_site, self.observed_direction).plus(
            self.observer_state.position
        )
        return PositionConvert.itrf.to_gcrf(itrf_ob, self.epoch())


class LiveOpticalObservation:
    def __init__(self, ob_dict):
        """interface used to access data stored in live observations

        :param ob_dict: single return from a Query JSON
        :type ob_dict: dict
        """
        self.epoch: Epoch = Epoch.from_udl_string(ob_dict["obTime"])
        self.sat_id: str = ob_dict.get("origObjectId", ob_dict.get("idOnOrbit", "UNKNOWN"))
        self.observer_id: str = ob_dict.get("origSensorId", ob_dict.get("sensorId", "NOT PROVIDED"))
        self.azimuth: float = radians(ob_dict.get("azimuth", 0))
        self.elevation: float = radians(ob_dict.get("elevation", 0))
        self.observer_lla: LLA = LLA(radians(ob_dict["senlat"]), radians(ob_dict["senlon"]), ob_dict["senalt"])
        self.observer_eci: Vector3D = Vector3D(ob_dict.get("senx", 0), ob_dict.get("seny"), ob_dict.get("senz"))
        self.right_ascension: float = radians(ob_dict["ra"])
        self.declination: float = radians(ob_dict["declination"])
        self.visual_magnitude: float = ob_dict.get("mag", 0)
        self.source: str = ob_dict["source"]
        self.mode: str = ob_dict["dataMode"]
        self.equatorial_phase_angle: float = ob_dict.get("solarEqPhaseAngle", 0)
        self.solar_declination_angle: float = ob_dict.get("solarDecAngle", 0)

    def get_observer_itrf(self) -> ITRF:
        return ITRF.from_fixed(self.epoch, PositionConvert.lla.to_itrf(self.observer_lla))

    def get_clos(self, tgt_gcrf: GCRF) -> float:
        observed = GroundObservation.from_angles_and_range(
            self.get_observer_itrf(), self.azimuth, self.elevation, 1, 0, 0
        )
        observed.observed_direction
        mat = ENZ.matrix(self.observer_lla.longitude, self.observer_lla.latitude)
        tgt_itrf = PositionConvert.gcrf.to_itrf(tgt_gcrf.position, tgt_gcrf.epoch)
        expected = mat.multiply_vector(tgt_itrf.minus(self.get_observer_itrf().position))
        return expected.magnitude() * expected.angle(observed.observed_direction)

    def csv_headers(self) -> str:
        return ",".join(
            [
                "UTC_EPOCH",
                "SOURCE",
                "TARGET_ID",
                "AZ",
                "EL",
                "RA",
                "DEC",
                "SOLAR_EQ_PHASE",
                "SOLAR_DEC_PHASE",
                "VISMAG",
                "SENSOR_ID",
                "SENSOR_LAT",
                "SENSOR_LON",
                "SENSOR_ALT",
            ]
        )

    def to_csv_format(self) -> str:
        return ",".join(
            [
                self.epoch.to_udl_string(),
                self.source,
                self.sat_id,
                str(self.azimuth),
                str(self.elevation),
                str(self.right_ascension),
                str(self.declination),
                str(self.equatorial_phase_angle),
                str(self.solar_declination_angle),
                str(self.visual_magnitude),
                self.observer_id,
                str(self.observer_lla.latitude),
                str(self.observer_lla.longitude),
                str(self.observer_lla.altitude),
            ]
        )


class LiveOpticalSet:
    def __init__(self):
        self.list = []
        self.sensors = {}
        self.targets = {}
        self.sources = {}
        self.modes = {}
        self.total = 0

    def publish_set(self, filename: str, write_mode: str) -> None:

        with open(filename, write_mode) as f:
            for ob in self.list:
                f.write("".join([ob.to_csv_format(), "\n"]))

    def process_observation(self, ob: LiveOpticalObservation) -> None:
        self.list.append(ob)
        self.total += 1
        if self.sensors.get(ob.observer_id) is None:
            self.sensors[ob.observer_id] = []
        if self.targets.get(ob.sat_id) is None:
            self.targets[ob.sat_id] = []
        if self.sources.get(ob.source) is None:
            self.sources[ob.source] = []
        if self.modes.get(ob.mode) is None:
            self.modes[ob.mode] = []

        self.sensors[ob.observer_id].append(ob)
        self.targets[ob.sat_id].append(ob)
        self.sources[ob.source].append(ob)
        self.modes[ob.mode].append(ob)

    def get_observers(self):
        return self.sensors.keys()

    def get_targets(self):
        return self.targets.keys()

    def get_sources(self):
        return self.sources.keys()



import unittest

from pysmad.coordinates.positions import PositionConvert
from pysmad.math.linalg import Vector3D
from pysmad.time import Epoch


class TestPositionConvertGCRF(unittest.TestCase):
    START_POS: Vector3D = Vector3D(10000, 40000, -5000)
    EPOCH = Epoch.from_gregorian(2021, 12, 25, 4, 43, 51.608)

    def test_to_itrf(self):
        itrf: Vector3D = PositionConvert.gcrf.to_itrf(self.START_POS, self.EPOCH)
        self.assertAlmostEqual(itrf.x, 1173.544602365, 0)
        self.assertAlmostEqual(itrf.y, -41216.97127606, 0)
        self.assertAlmostEqual(itrf.z, -4978.360362079, 0)



import unittest

from pysmad.coordinates.states import GCRF
from pysmad.math.linalg import Vector3D
from pysmad.time import Epoch


class TestGCRFstate(unittest.TestCase):

    EPOCH = Epoch.from_gregorian(2021, 12, 25, 4, 43, 51.608)
    STATE: GCRF = GCRF(EPOCH, Vector3D(10000, 40000, -5000), Vector3D(0, 0, 0))
    EARTH_SURFACE: GCRF = GCRF(EPOCH, Vector3D(6378, 0, 0), Vector3D(0, 0, 0))

    def test_total_acceleration_from_earth(self):
        a = self.EARTH_SURFACE.acceleration_from_gravity().plus(self.EARTH_SURFACE.acceleration_from_earth())
        self.assertAlmostEqual(0.009814696076649412, a.magnitude())

    def test_acceleration_from_gravity(self):
        a = self.EARTH_SURFACE.acceleration_from_gravity()
        self.assertAlmostEqual(1.5990836818112354e-05, a.magnitude(), 15)

    def test_acceleration_from_earth(self):
        a = self.EARTH_SURFACE.acceleration_from_earth()
        self.assertAlmostEqual(0.00979870641977297, a.magnitude(), 15)

    def test_acceleration_from_sun(self):
        a = self.STATE.acceleration_from_sun()
        self.assertAlmostEqual(3.00696687920662e-9, a.magnitude(), 15)

    def test_acceleration_from_moon(self):
        a = self.STATE.acceleration_from_moon()
        self.assertAlmostEqual(3.4064259486775476e-9, a.magnitude(), 15)



from math import cos, radians, sin, sqrt
from typing import List

from pysmad.math.constants import SECONDS_IN_SIDEREAL_DAY
from pysmad.math.functions import Conversions, EquationsOfMotion
from pysmad.math.linalg import Matrix3D, Vector3D
from pysmad.time import Epoch


class Earth:
    """Class used to store Earth properties"""

    #: normalized c coefficients used for geopotential calculation
    C: List[List[float]] = [
        [1],
        [0, 0],
        [-0.484165143790815e-3 / sqrt(0.2), -0.206615509074176e-9 / sqrt(0.6), 0.243938357328313e-5 / sqrt(2.4)],
        [
            0.957161207093473e-6 / sqrt(1 / 7),
            0.203046201047864e-5 / sqrt(6 / 7),
            0.904787894809528e-6 / sqrt(60 / 7),
            0.721321757121568e-6 / sqrt(360 / 7),
        ],
        [
            0.539965866638991e-6 / sqrt(24 / 216),
            -0.536157389388867e-6 / sqrt(10 / 9),
            0.350501623962649e-6 / sqrt(20),
            0.990856766672321e-6 / sqrt(280),
            -0.188519633023033e-6 / sqrt(2240),
        ],
    ]

    #: normalized s coefficients used for geopotential calculation
    S: List[List[float]] = [
        [0],
        [0, 0],
        [0, 0.138441389137979e-8 / sqrt(0.6), -0.140027370385934e-5 / sqrt(2.4)],
        [
            0,
            0.248200415856872e-6 / sqrt(6 / 7),
            -0.619005475177618e-6 / sqrt(60 / 7),
            0.141434926192941e-5 / sqrt(360 / 7),
        ],
        [
            0,
            -0.473567346518086e-6 / sqrt(10 / 9),
            0.662480026275829e-6 / sqrt(20),
            -0.200956723567452e-6 / sqrt(280),
            0.308803882149194e-6 / sqrt(2240),
        ],
    ]

    #: the number of zonal and tesseral terms to be used in the gravity calculation
    DEGREE_AND_ORDER: int = len(S)

    #: G*M given in km^3/s^2
    MU: float = 398600.4418

    #: distance from earth center to surface at the equator in km
    RADIUS: float = 6378.137

    #: value defining the ellipsoid of an oblate earth
    FLATTENING: float = 1 / 298.2572235

    #: inclination of ecliptic relative to earth equator in radians
    OBLIQUITY_OF_ECLIPTIC: float = radians(23.43929111)

    #: boolean identifying if gravity will be modeled as a point-source or with non-spherical methods
    USE_GEODETIC_MODEL = True

    @staticmethod
    def obliquity_of_ecliptic_at_epoch(epoch: Epoch) -> float:
        """calculate the obliquity of ecliptic (epsilon) at a given epoch

        :param epoch: time of interest
        :type epoch: Epoch
        :return: true-of-date epsilon
        :rtype: float
        """
        # Equation 5.39
        t = epoch.julian_centuries_past_j2000()

        a = Conversions.dms_to_radians(0, 0, 46.815)
        b = Conversions.dms_to_radians(0, 0, 0.00059)
        c = Conversions.dms_to_radians(0, 0, 0.001813)

        # Equation 5.42
        return Earth.OBLIQUITY_OF_ECLIPTIC - a * t - b * t * t + c * t * t * t

    @staticmethod
    def geo_radius() -> float:
        """calculate the radius of an orbit with a period equal to that of Earth's rotation

        :return: geosynchronous radius in km
        :rtype: float
        """
        return EquationsOfMotion.A.from_mu_tau(Earth.MU, SECONDS_IN_SIDEREAL_DAY)

    @staticmethod
    def rotation(epoch: Epoch) -> Matrix3D:
        """creates a matrix that can be used to transform an itrf position to tod and vice versa

        :param epoch: valid time of the state
        :type epoch: Epoch
        :return: transformation matrix
        :rtype: Matrix3D
        """
        d: float = epoch.julian_value() - Epoch.J2000_JULIAN_DATE
        arg1: float = radians(125.0 - 0.05295 * d)
        arg2: float = radians(200.9 + 1.97129 * d)
        a: float = radians(-0.0048)
        b: float = radians(0.0004)
        dpsi: float = a * sin(arg1) - b * sin(arg2)
        eps: float = Earth.obliquity_of_ecliptic_at_epoch(epoch)
        gmst: float = epoch.greenwich_hour_angle()
        gast: float = dpsi * cos(eps) + gmst
        return Vector3D.rotation_matrix(Vector3D(0, 0, 1), -gast)

    @staticmethod
    def precession(epoch: Epoch) -> Matrix3D:
        """creates a matrix that can be used to transform a tod position to mod and vice versa

        :param epoch: valid time of the state
        :type epoch: Epoch
        :return: transformation matrix
        :rtype: Matrix3D
        """
        t: float = epoch.julian_centuries_past_j2000()
        a: float = Conversions.dms_to_radians(0, 0, 2306.2181)
        b: float = Conversions.dms_to_radians(0, 0, 0.30188)
        c: float = Conversions.dms_to_radians(0, 0, 0.017998)
        x: float = a * t + b * t * t + c * t * t * t

        a = Conversions.dms_to_radians(0, 0, 2004.3109)
        b = Conversions.dms_to_radians(0, 0, 0.42665)
        c = Conversions.dms_to_radians(0, 0, 0.041833)
        y: float = a * t - b * t * t - c * t * t * t

        a = Conversions.dms_to_radians(0, 0, 0.7928)
        b = Conversions.dms_to_radians(0, 0, 0.000205)
        z: float = x + a * t * t + b * t * t * t

        sz = sin(z)
        sy = sin(y)
        sx = sin(x)
        cz = cos(z)
        cy = cos(y)
        cx = cos(x)

        return Matrix3D(
            Vector3D(-sz * sx + cz * cy * cx, -sz * cx - cz * cy * sx, -cz * sy),
            Vector3D(cz * sx + sz * cy * cx, cz * cx - sz * cy * sx, -sz * sy),
            Vector3D(sy * cx, -sy * sx, cy),
        )

    @staticmethod
    def nutation(epoch: Epoch) -> Matrix3D:
        """creates a matrix that can be used to transform a mod position to gcrf and vice versa

        :param epoch: valid time of the state
        :type epoch: Epoch
        :return: transformation matrix
        :rtype: Matrix3D
        """
        d = epoch.julian_value() - Epoch.J2000_JULIAN_DATE
        arg1 = radians(125.0 - 0.05295 * d)
        arg2 = radians(200.9 + 1.97129 * d)
        a = radians(-0.0048)
        b = radians(0.0004)
        dpsi = a * sin(arg1) - b * sin(arg2)
        a = radians(0.0026)
        b = radians(0.0002)
        deps = a * cos(arg1) + b * cos(arg2)
        eps = Earth.obliquity_of_ecliptic_at_epoch(epoch)

        ce = cos(eps)
        se = sin(eps)

        return Matrix3D(
            Vector3D(1.0, -dpsi * ce, -dpsi * se),
            Vector3D(dpsi * ce, 1.0, -deps),
            Vector3D(dpsi * se, deps, 1.0),
        )


class Sun:
    """class used to store properties of the sun for a force model"""

    #: RAAN + argument of periapsis in radians
    W_PLUS_W = radians(282.94)

    #: cosine of obliquity of ecliptic
    COS_OBLIQUITY = cos(Earth.OBLIQUITY_OF_ECLIPTIC)

    #: sine of obliquity of ecliptic
    SIN_OBLIQUITY = sin(Earth.OBLIQUITY_OF_ECLIPTIC)

    #: G*M in km^3/s^
    MU = 1.327124400419e11

    #: Estimate of srp
    P = 4.56e-6

    #: Distance to earth in km
    AU = 149597870.691

    @staticmethod
    def get_position(epoch: Epoch) -> Vector3D:
        """calculate the ECI position at a given epoch

        :param epoch: time of calculated position vector
        :type epoch: Epoch
        :return: ECI position in km
        :rtype: Vector3D
        """
        a = Conversions.dms_to_radians(0, 0, 6892)
        b = Conversions.dms_to_radians(0, 0, 72)
        t = epoch.julian_centuries_past_j2000()

        ma = radians(357.5256 + 35999.049 * t)

        sma = sin(ma)
        cma = cos(ma)
        c2ma = cos(2 * ma)

        lam = Sun.W_PLUS_W + ma + a * sma + b * 2 * sma * cma
        r = (149.619 - 2.499 * cma - 0.021 * c2ma) * 1e6

        x = r * cos(lam)

        slam = sin(lam)

        y = r * slam * Sun.COS_OBLIQUITY
        z = r * slam * Sun.SIN_OBLIQUITY

        return Vector3D(x, y, z)


class Moon:
    """class used to store properties of the moon to be used in force modeling"""

    #: G*M in km^3/s^2
    MU = 4902.800305555

    #: distance from center of moon to surface in km
    RADIUS = 1737.4000

    @staticmethod
    def get_position(epoch: Epoch) -> Vector3D:
        """calculate ECI position of moon

        :param epoch: time of calculated position vector
        :type epoch: Epoch
        :return: ECI position in km
        :rtype: Vector3D
        """
        # Equation 3.47
        t = epoch.julian_centuries_past_j2000()
        l0 = radians(218.31617 + 481267.88088 * t - 1.3972 * t)
        l = radians(134.96292 + 477198.86753 * t)
        lp = radians(357.52543 + 35999.04944 * t)
        f = radians(93.27283 + 483202.01873 * t)
        d = radians(297.85027 + 445267.11135 * t)

        # Auxiliary variables to store the angles defined in seconds in equation 3.48
        a0 = Conversions.dms_to_radians(0, 0, 22640.0)
        a1 = Conversions.dms_to_radians(0, 0, 769.0)
        a2 = Conversions.dms_to_radians(0, 0, 4586.0)
        a3 = Conversions.dms_to_radians(0, 0, 2370.0)
        a4 = Conversions.dms_to_radians(0, 0, 668.0)
        a5 = Conversions.dms_to_radians(0, 0, 412.0)
        a6 = Conversions.dms_to_radians(0, 0, 212.0)
        a7 = Conversions.dms_to_radians(0, 0, 206.0)
        a8 = Conversions.dms_to_radians(0, 0, 192.0)
        a9 = Conversions.dms_to_radians(0, 0, 165.0)
        a10 = Conversions.dms_to_radians(0, 0, 148.0)
        a11 = Conversions.dms_to_radians(0, 0, 125.0)
        a12 = Conversions.dms_to_radians(0, 0, 110.0)
        a13 = Conversions.dms_to_radians(0, 0, 55.0)

        # Equation 3.48
        lam = (
            l0
            + a0 * sin(l)
            + a1 * sin(2 * l)
            - a2 * sin(l - 2.0 * d)
            + a3 * sin(2.0 * d)
            - a4 * sin(lp)
            - a5 * sin(2.0 * f)
            - a6 * sin(2.0 * l - 2.0 * d)
            - a7 * sin(l + lp - 2.0 * d)
            + a8 * sin(l + 2.0 * d)
            - a9 * sin(lp - 2.0 * d)
            + a10 * sin(l - lp)
            - a11 * sin(d)
            - a12 * sin(l + lp)
            - a13 * sin(2.0 * f - 2.0 * d)
        )

        # Redefined variables for angles in equation 3.49
        a0 = Conversions.dms_to_radians(0, 0, 18520.0)
        a1 = Conversions.dms_to_radians(0, 0, 412.0)
        a2 = Conversions.dms_to_radians(0, 0, 541.0)
        a3 = Conversions.dms_to_radians(0, 0, 526.0)
        a4 = Conversions.dms_to_radians(0, 0, 44.0)
        a5 = Conversions.dms_to_radians(0, 0, 31.0)
        a6 = Conversions.dms_to_radians(0, 0, 25.0)
        a7 = Conversions.dms_to_radians(0, 0, 23.0)
        a8 = Conversions.dms_to_radians(0, 0, 21.0)
        a9 = Conversions.dms_to_radians(0, 0, 11.0)

        # Equation 3.49
        beta = (
            a0 * sin(f + lam - l0 + a1 * sin(2.0 * f + a2 * sin(lp)))
            - a3 * sin(f - 2.0 * d)
            + a4 * sin(l + f - 2.0 * d)
            - a5 * sin(-l + f - 2.0 * d)
            - a6 * sin(-2.0 * l + f)
            - a7 * sin(lp + f - 2.0 * d)
            + a8 * sin(-l + f)
            + a9 * sin(-lp + f - 2.0 * d)
        )

        # Equation 3.50
        r = (
            385000.0
            - 20905.0 * cos(l)
            - 3699.0 * cos(2.0 * d - l)
            - 2956.0 * cos(2.0 * d)
            - 570.0 * cos(2.0 * l)
            + 246.0 * cos(2.0 * l - 2.0 * d)
            - 205.0 * cos(lp - 2.0 * d)
            - 171.0 * cos(l + 2.0 * d)
            - 152.0 * cos(l + lp - 2.0 * d)
        )

        # Equation 3.51
        x = r * cos(lam) * cos(beta)
        y = r * sin(lam) * cos(beta)
        z = r * sin(beta)

        eps = Earth.OBLIQUITY_OF_ECLIPTIC

        return Vector3D(x, y, z).rotation_about_axis(Vector3D(1, 0, 0), eps)



import unittest

from pysmad.coordinates.states import GCRF
from pysmad.math.linalg import Vector3D
from pysmad.propagators.inertial import RK4
from pysmad.time import Epoch


class TestRK4(unittest.TestCase):

    EPOCH: Epoch = Epoch.from_gregorian(2022, 12, 20, 0, 1, 9.184)
    POSITION: Vector3D = Vector3D(42164, 0, 0)
    VELOCITY: Vector3D = Vector3D(0, 3.07375, 0)
    STATE: GCRF = GCRF(EPOCH, POSITION, VELOCITY)
    PROPAGATOR: RK4 = RK4(STATE)

    def test_step_to_epoch(self):
        end_epoch: Epoch = Epoch.from_gregorian(2022, 12, 20, 19, 1, 9.184)
        self.PROPAGATOR.step_to_epoch(end_epoch)
        self.assertAlmostEqual(self.PROPAGATOR.state.position.x, 11701.163084, -1)
        self.assertAlmostEqual(self.PROPAGATOR.state.position.y, -40487.016256, -1)
        self.assertAlmostEqual(self.PROPAGATOR.state.position.z, -2.099302, -1)
        self.assertAlmostEqual(self.PROPAGATOR.state.velocity.x, 2.954853, 4)
        self.assertAlmostEqual(self.PROPAGATOR.state.velocity.y, 0.851923, 4)
        self.assertAlmostEqual(self.PROPAGATOR.state.velocity.z, -0.000141, 4)



import unittest
from math import radians

from pysmad.bodies.terrestrial import GroundSite
from pysmad.coordinates.states import ITRF
from pysmad.estimation.obs import GroundObservation
from pysmad.math.linalg import Vector3D
from pysmad.time import Epoch


class TestGroundObservation(unittest.TestCase):
    SITE: GroundSite = GroundSite.from_itrf_position(Vector3D(1344.143, 6068.601, 1429.311))

    def test_ijk_position(self):
        state_1: ITRF = ITRF.from_fixed(Epoch.from_gregorian(1999, 4, 2, 0, 31, 9.184), self.SITE.itrf_position)
        ob_1: GroundObservation = GroundObservation.from_angles_and_range(
            state_1, radians(132.67), radians(32.44), 16945.45, 0, 0
        )
        ijk_1: Vector3D = ob_1.ijk_position()
        self.assertAlmostEqual(ijk_1.x, 11959.977742417275)
        self.assertAlmostEqual(ijk_1.y, -16289.477980186362)
        self.assertAlmostEqual(ijk_1.z, -5963.827409736864)

    def test_site_itrf_position(self):
        self.assertAlmostEqual(self.SITE.itrf_position.x, 1344.143)
        self.assertAlmostEqual(self.SITE.itrf_position.y, 6068.601)
        self.assertAlmostEqual(self.SITE.itrf_position.z, 1429.311)

    def test_site_ijk_position(self):
        gmst: float = Epoch.from_gregorian(1999, 4, 2, 0, 31, 9.184).greenwich_hour_angle()
        ijk: Vector3D = self.SITE.itrf_position.rotation_about_axis(Vector3D(0, 0, 1), gmst)
        self.assertAlmostEqual(ijk.x, 534.388240046999)
        self.assertAlmostEqual(ijk.y, -6192.662408903832)
        self.assertAlmostEqual(ijk.z, 1429.3110)



from pysmad.bodies.artificial import Spacecraft
from pysmad.coordinates.positions import ENZ, LLA, PositionConvert
from pysmad.coordinates.states import ITRF
from pysmad.estimation.obs import GroundObservation
from pysmad.math.linalg import Matrix3D, Vector3D


class GroundSite:
    def __init__(self, lla: LLA) -> None:
        """used for operations that require modeling of a terrestrial location
        :param lla: state that holds the latitude, longitude, and altitude of the ground station
        :type lla: LLA
        """
        #: geodetic cartesian coordinates of the ground station in km
        self.itrf_position: Vector3D = PositionConvert.lla.to_itrf(lla)

        #: geodetic latitude in radians
        self.latitude: float = lla.latitude

        #: longitude in radians
        self.longitude: float = lla.longitude

        #: matrix used to perform transformations to and from enz/itrf
        self.enz_matrix: Matrix3D = ENZ.matrix(lla.longitude, lla.latitude)

    @classmethod
    def from_itrf_position(cls, itrf: Vector3D) -> "GroundSite":
        """create a groundsite from cartesian coordinates

        :param itrf: earth-fixed position of the groundsite
        :type itrf: Vector3D
        :return: object stationed at the input earth-fixed vector
        :rtype: GroundSite
        """
        return cls(PositionConvert.itrf.to_lla(itrf))

    def enz_position(self, obj_itrf: Vector3D) -> Vector3D:
        """calculates the east-north-zenith coordinates of the argument position

        :param obj_itrf: itrf position of the object of interest
        :type obj_itrf: Vector3D
        :return: transformation of the argument itrf position to the east-north-zenith frame
        :rtype: Vector3D
        """
        return self.enz_matrix.multiply_vector(obj_itrf.minus(self.itrf_position))

    def angles_and_range(self, target: Spacecraft) -> GroundObservation:
        """calculate the topo-centric angles and range to the argument spacecraft

        :param target: spacecraft being observed
        :type target: Spacecraft
        :return: azimuth, elevation, and range to the spacecraft from the ground site
        :rtype: GroundObservation
        """

        return GroundObservation(
            ITRF.from_fixed(target.current_epoch(), self.itrf_position),
            self.enz_position(PositionConvert.gcrf.to_itrf(target.position(), target.current_epoch())),
            0,
            0,
        )



from math import atan2, cos, pi, radians, sin, sqrt
from typing import List

from pysmad.math.constants import HOURS_IN_DAY, MINUTES_IN_DAY, MINUTES_IN_HOUR, SECONDS_IN_DAY, SECONDS_IN_HOUR
from pysmad.math.linalg import Vector3D


class Conversions:
    """class used to perform various unit conversions

    :return: defined by each static method
    :rtype: defined by each static method
    """

    @staticmethod
    def hms_to_decimal_day(hr: float, m: float, s: float) -> float:
        """calculate the float value of the day given an hour, minute, second representation

        :param hr: hour of day
        :type hr: float
        :param m: minute of day
        :type m: float
        :param s: second of day
        :type s: float
        :return: the time in days
        :rtype: float
        """
        return hr / HOURS_IN_DAY + m / MINUTES_IN_DAY + s / SECONDS_IN_DAY

    @staticmethod
    def dms_to_radians(d: float, m: float, s: float) -> float:
        """calculate an angle in radians that has been defined in degrees, minutes, and seconds

        :param d: degrees in angle
        :type d: float
        :param m: minute of angle
        :type m: float
        :param s: second of angle
        :type s: float
        :return: angle in radians
        :rtype: float
        """
        return radians(d + m / MINUTES_IN_HOUR + s / SECONDS_IN_HOUR)


def sign(num: float) -> float:
    """function to determine if a value is positive or negative

    :param num: expression to be signed
    :type num: float
    :return: 1 if positive -1 if negative 0 if neither
    :rtype: float
    """
    val = 0
    if num > 0:
        val = 1
    elif num < 0:
        val = -1
    return val


class LegendrePolynomial:
    def __init__(self, phi: float) -> None:
        """stores the explicit solution to normalized legendre polynomials used in the geopotential calculations

        :param phi: geodetic latitude in radians
        :type phi: float
        """
        cos_phi: float = cos(phi)
        sin_phi: float = sin(phi)
        cos_phi_squared: float = cos_phi * cos_phi
        sin_phi_squared: float = sin_phi * sin_phi

        #: the polynomial list of lists with indices n, m
        self.p: List[List[float]] = [
            [1, 0],
            [sin_phi, cos_phi, 0],
            [
                (3 * sin_phi_squared - 1) * 0.5,
                3 * sin_phi * cos_phi,
                3 * cos_phi_squared,
                0,
            ],
            [
                sin_phi * (5 * sin_phi_squared - 3) * 0.5,
                (15 * sin_phi_squared - 3) * cos_phi * 0.5,
                15 * sin_phi * cos_phi_squared,
                15 * cos_phi_squared * cos_phi,
                0,
            ],
            [
                0.125 * (35 * sin_phi_squared * sin_phi_squared - 30 * sin_phi_squared + 3),
                2.5 * (7 * sin_phi_squared * sin_phi - 3 * sin_phi) * cos_phi,
                (7 * sin_phi_squared - 1) * cos_phi_squared * 7.5,
                105 * cos_phi * cos_phi_squared * sin_phi,
                105 * cos_phi_squared * cos_phi_squared,
                0,
            ],
        ]


class Eccentricity:
    r"""Class used to solve eccentricity of an ellipse

    .. note::

       eccentricity will commonly be referenced as :math:`e` in documentation"""

    @staticmethod
    def from_a_c(a: float, c: float) -> float:
        r"""calculate eccentricity using equation 1-2 in :ref:`vallado`

        .. math::

           e = \frac{c}{a}

        :param a: semi-major axis in :math:`km`
        :type a: float
        :param c: half the distance between focii in :math:`km`
        :type c: float
        :return: eccentricity
        :rtype: float

        .. note::

           This method works any time :math:`a` and :math:`c` are input with common units,
           but :math:`km` is the preferred unit throughout this package to ensure computational
           consistency and accuracy.
        """
        return c / a

    @staticmethod
    def from_a_b(a: float, b: float) -> float:
        r"""calculate eccentricity using equation 1-6 in :ref:`vallado`

        .. math::

           e = \frac{\sqrt{a^2-b^2}}{a}

        :param a: semi-major axis in :math:`km`
        :type a: float
        :param b: semi-minor axis in :math:`km`
        :type b: float
        :return: eccentricity
        :rtype: float

        .. note::

           This method works any time :math:`a` and :math:`b` are input with common units,
           but :math:`km` is the preferred unit throughout this package to ensure computational
           consistency and accuracy.
        """
        return sqrt(a * a - b * b) / a

    @staticmethod
    def from_a_p(a: float, p: float) -> float:
        r"""calculate eccentricity using equation 2.62 in :ref:`orbits`

        .. math::

           e = \sqrt{1 - \frac{p}{a}}

        :param a: semi-major axis in :math:`km`
        :type a: float
        :param p: semi-parameter in :math:`km`
        :type p: float
        :return: eccentricity
        :rtype: float

        .. note::

           This method works any time :math:`a` and :math:`p` are input with common units,
           but :math:`km` is the preferred unit throughout this package to ensure computational
           consistency and accuracy.
        """
        return sqrt(1 - p / a)


class Flattening:
    r"""class used to solve flattening of an ellipse

    .. note::

       flattening will commonly be referenced as :math:`f` in documentation"""

    @staticmethod
    def from_a_b(a: float, b: float) -> float:
        r"""calculate flattening using equation 1-3 in :ref:`vallado`

        .. math::

           f = \frac{a-b}{a}

        :param a: semi-major axis in :math:`km`
        :type a: float
        :param b: semi-minor axis in :math:`km`
        :type b: float
        :return: flattening
        :rtype: float
        """
        return (a - b) / a


class SemiMajorAxis:
    """class used to solve semi-major axis of an ellipse

    .. note::

       the semi-major axis will commonly be referenced as :math:`a` in documentation
    """

    @staticmethod
    def from_mu_n(mu: float, n: float) -> float:
        r"""calculate semi-major axis using equation 1-29 in :ref:`vallado`

        .. math::

           a = \sqrt[3]{\frac{\mu}{n^2}}

        :param mu: gravitational constant time mass of central body combined units of :math:`\frac{km^3}{s^2}`
        :type mu: float
        :param n: mean motion in :math:`\frac{rad}{s}`
        :type n: float
        :return: semi-major axis in :math:`km`
        :rtype: float
        """
        return (mu / (n * n)) ** (1 / 3)

    @staticmethod
    def from_mu_tau(mu: float, tau: float) -> float:
        r"""calculate semi-major axis using equation 1-27 in :ref:`vallado`

        .. math::

           n = \frac{2\pi}{\tau} = \sqrt{\frac{\mu}{a^3}}

        :param mu: gravitational constant time mass of central body combined units of :math:`\frac{km^3}{s^2}`
        :type mu: float
        :param tau: period in :math:`s`
        :type tau: float
        :return: semi-major axis in :math:`km`
        :rtype: float
        """
        base: float = tau / (2 * pi)
        return (mu * base * base) ** (1 / 3)

    @staticmethod
    def from_mu_r_v(mu: float, r: float, v: float) -> float:
        r"""calculate the semi-major axis in km using equation 1-31 in :ref:`vallado`

        .. math::

           v = \sqrt{\frac{2\mu}{r}+\frac{\mu}{a}}

        :param mu: gravitational constant time mass of central body combined units of :math:`\frac{km^3}{s^2}`
        :type mu: float
        :param r: magnitude of the position vector in :math:`km`
        :type r: float
        :param v: magnitude of the velocity vector in :math:`\frac{km}{s}`
        :type v: float
        :return: semi-major axis in :math:`km`
        :rtype: float
        """
        return 1 / (2 / r - v * v / mu)


class SemiMinorAxis:
    """class used to solve semi-minor axis of an ellipse

    .. note::

    the semi-minor axis will commonly be referenced as :math:`b` in documentation"""

    @staticmethod
    def from_a_e(a: float, e: float) -> float:
        r"""calculate semi-minor axis using equation 1-4 in :ref:`vallado`

        .. math::

           b = a\sqrt{1-e^2}

        :param a: semi-major axis in :math:`km`
        :type a: float
        :param e: eccentricity
        :type e: float
        :return: semi-minor axis in :math:`km`
        :rtype: float
        """
        return a * sqrt(1 - e * e)


class SemiParameter:
    """class used to solve the semi-parameter of an ellipse

    .. note::

    the semi-parameter will commonly be referenced as :math:`p` in documentation"""

    @staticmethod
    def from_a_b(a: float, b: float) -> float:
        r"""calculate the semi-parameter using equation 1-9 in :ref:`vallado`

        .. math::

           p = \frac{b^2}{a}

        :param a: semi-major axis in :math:`km`
        :type a: float
        :param b: semi-minor axis in :math:`km`
        :type b: float
        :return: semi-parameter in :math:`km`
        :rtype: float
        """
        return b * b / a

    @staticmethod
    def from_a_e(a: float, e: float) -> float:
        r"""calculate the semi-parameter using equation 1-10 in :ref:`vallado`

        .. math::

           p = a(1-e^2)

        :param a: semi-major axis in :math:`km`
        :type a: float
        :param e: eccentricity
        :type e: float
        :return: semi-parameter in :math:`km`
        :rtype: float
        """
        return a * (1 - e * e)

    @staticmethod
    def from_mu_h(mu: float, h: float) -> float:
        r"""calculate the semi-parameter using equation 1-19 in :ref:`vallado`

        .. math::

           p = \frac{h^2}{\mu}

        :param mu: gravitational constant time mass of central body combined units of :math:`\frac{km^3}{s^2}`
        :type mu: float
        :param h: areal velocity magnitude in :math:`\frac{km^2}{s}`
        :type h: float
        :return: semi-parameter in :math:`km`
        :rtype: float
        """
        return h * h / mu


class ArealVelocity:
    r"""class used to calculate areal velocities of an orbit

    .. note::
       The term areal velocity is used synonymously for the momentum of an orbit
       and will commonly be referenced as :math:`h` in documentation

    """

    @staticmethod
    def from_r_v_phi(r: float, v: float, phi: float) -> float:
        r"""calculate the areal velocity using equation 1-16 in :ref:`vallado`

        .. math::

           h = rv\cos{(\phi_{fpa})}

        :param r: magnitude of the position vector in :math:`km`
        :type r: float
        :param v: magnitude of the velocity vector in :math:`\frac{km}{s}`
        :type v: float
        :param phi: flight path angle :math:`\frac{\pi}{2} - \measuredangle\vec{r}\vec{v}`
        :type phi: float
        :return: areal velocity in :math:`\frac{km^2}{s}`
        :rtype: float
        """
        return r * v * cos(phi)

    @staticmethod
    def from_mu_p(mu: float, p: float) -> float:
        r"""calculate the areal velocity using equation 1-19 in :ref:`vallado`

        .. math::

           h = \sqrt{{\mu}p}

        :param mu: gravitational constant time mass of central body combined units of :math:`\frac{km^3}{s^2}`
        :type mu: float
        :param p: semi-parameter in :math:`km`
        :type p: float
        :return: areal velocity in :math:`\frac{km^2}{s}`
        :rtype: float
        """
        return sqrt(mu * p)

    @staticmethod
    def from_r_v(r: Vector3D, v: Vector3D) -> Vector3D:
        r"""calculate the momentum vector using equation 1-15 in :ref:`vallado`

        .. math::

           \vec{h} = \vec{r} \times \vec{v}

        :param r: position vector in :math:`km`
        :type r: Vector3D
        :param v: velocity vector in :math:`\frac{km}{s}`
        :type v: Vector3D
        :return: areal velocity vector in :math:`\frac{km^2}{s}`
        :rtype: Vector3D
        """
        return r.cross(v)


class RightAscensionNode:
    r"""static class used to solve right ascension of ascending node (RAAN)

    .. note::

       RAAN will commonly be referenced as :math:`\Omega` in documentation
    """

    @staticmethod
    def from_w(w: Vector3D) -> float:
        r"""calculate the right ascension of the ascending node

        .. math::

           \Omega = \arctan{\left(-\frac{\vec{w}_{x}}{\vec{w}_{y}}\right)}

        :param w: normalized momentum vector in :math:`\frac{km^2}{s}`
        :type w: Vector3D
        :return: right ascension of the ascending node in :math:`rad` where :math:`0 \leq \Omega < 2\pi`
        :rtype: float

        .. todo::

           document equation reference from Satellite Orbits
        """
        raan: float = atan2(w.x, -w.y)
        if raan < 0:
            raan += 2 * pi
        return raan


class Inclination:
    r"""static class used to solve inclination

    .. note::

       inclination will commonly be referenced as :math:`i` in documentation
    """

    @staticmethod
    def from_w(w: Vector3D) -> float:
        r"""calculate the inclination

        .. math::

           i = \arctan{
                \left(
                    \frac{
                        \sqrt{
                            \vec{w}_{x}^2 + \vec{w}_{y}^2
                        }
                    }{
                        \vec{w}_{z}
                    }
                \right)
            }

        :param w: normalized momentum vector :math:`\hat{h}` in :math:`\frac{km^2}{s}`
        :type w: Vector3D
        :return: inclination in :math:`rad` where :math:`0 \leq i < \pi`
        :rtype: float

        .. todo::

           document equation reference from Satellite Orbits
        """
        return atan2(sqrt(w.x * w.x + w.y * w.y), w.z)


class Radius:
    r"""static class used to solve the radius of an orbit

    .. note::

       radius will commonly be referenced as :math:`r` or :math:`R` in documentation
    """

    @staticmethod
    def from_p_e_nu(p: float, e: float, nu: float) -> float:
        r"""calculate the radius of an orbit using equation 1-24 in :ref:`vallado`

        .. math::

           r = \frac{p}{1+e\cos{(\nu)}}

        :param p: semi-parameter in :math:`km`
        :type p: float
        :param e: eccentricity
        :type e: float
        :param nu: true anomaly in :math:`rads`
        :type nu: float
        :return: radius in :math:`km`
        :rtype: float
        """
        return p / (1 + e * cos(nu))


class SpecificMechanicalEnergy:
    r"""static class used to solve specific mechanical energy

    .. note::

       SME will commonly be referenced as :math:`\xi` in documentation
    """

    @staticmethod
    def from_mu_r_v(mu: float, r: float, v: float) -> float:
        r"""calculate the specific mechanical energy using equation 1-20 in :ref:`vallado`

        .. math::

           \xi = \frac{v^2}{2} - \frac{\mu}{r}

        :param mu: gravitational constant time mass of central body combined units of :math:`\frac{km^3}{s^2}`
        :type mu: float
        :param r: distance from center of central body in :math:`km`
        :type r: float
        :param v: magnitude of velocity vector in :math:`\frac{km}{s}`
        :type v: float
        :return: specific mechanical energy in :math:`\frac{km^2}{s^2}`
        :rtype: float
        """
        return v * v * 0.5 - mu / r

    @staticmethod
    def from_mu_a(mu: float, a: float) -> float:
        r"""calculate specific mechanical energy using equation 1-21 in :ref:`vallado`

        .. math::

           \xi = -\frac{\mu}{2a}

        :param mu: gravitational constant time mass of central body combined units of :math:`\frac{km^3}{s^2}`
        :type mu: float
        :param a: semi-major axis in :math:`km`
        :type a: float
        :return: specific mechanical energy :math:`\frac{km^2}{s^2}`
        :rtype: float
        """
        return -0.5 * mu / a


class VisVivaVelocity:
    r"""class used to calculate the velocity of an orbit

    .. note::

       velocity will commonly be referenced as :math:`v` in documentation
    """

    @staticmethod
    def from_a_mu_r(a: float, mu: float, r: float) -> float:
        r"""calculate magnitude of velocity using equation 1-22 in :ref:`vallado`

        .. math::

           v^2 = \mu\left(\frac{2}{r} - \frac{1}{a}\right)

        :param a: semi-major axis in :math:`km`
        :type a: float
        :param mu: gravitational constant time mass of central body combined units of :math:`\frac{km^3}{s^2}`
        :type mu: float
        :param r: distance from center of central body in :math:`km`
        :type r: float
        :return: velocity magnitude in :math:`\frac{km}{s}`
        :rtype: float
        """
        return sqrt(mu * (2 / r - 1 / a))

    @staticmethod
    def from_mu_r_xi(mu: float, r: float, xi: float) -> float:
        r"""calculate the magnitude of velocity using equation 1-30 in :ref:`vallado`

        .. math::
           v = \sqrt{2\left(\frac{\mu}{r}+\xi\right)}

        :param mu: gravitational constant time mass of central body combined units of :math:`\frac{km^3}{s^2}`
        :type mu: float
        :param r: distance from center of central body in :math:`km`
        :type r: float
        :param xi: specific mechanical energy in :math:`\frac{km^2}{s^2}`
        :type xi: float
        :return: velocity in :math:`\frac{km}{s}`
        :rtype: float
        """
        return sqrt(2 * (mu / r + xi))

    @staticmethod
    def from_mu_r_e_nu(mu: float, r: float, e: float, nu: float) -> float:
        r"""calculate the magnitude of velocity using equation 1-32 in :ref:`vallado`

        .. math::

           v = \sqrt{\frac{\mu}{r}\left(2 - \frac{1-e^2}{1+e\cos{\nu}}\right)}

        :param mu: gravitational constant time mass of central body combined units of :math:`\frac{km^3}{s^2}`
        :type mu: float
        :param r: distance from center of central body in :math:`km`
        :type r: float
        :param e: eccentricity
        :type e: float
        :param nu: true anomaly in :math:`rads`
        :type nu: float
        :return: velocity in :math:`\frac{km}{s}`
        :rtype: float
        """
        return sqrt((mu / r) * (2 - (1 - e * e) / (1 + e * cos(nu))))


class Period:
    r"""class used to calculate the period of an orbit

    .. note::

       period will commonly be referenced as :math:`\tau` in documentation"""

    @staticmethod
    def from_a_mu(a: float, mu: float) -> float:
        r"""calculate the period using equation 1-26 in :ref:`vallado`

        .. math::

           \tau = 2\pi\sqrt{\frac{a^3}{\mu}}

        :param a: semi-major axis in :math:`km`
        :type a: float
        :param mu: gravitational constant time mass of central body combined units of :math:`\frac{km^3}{s^2}`
        :type mu: float
        :return: period in :math:`s`
        :rtype: float
        """
        return 2 * pi * sqrt(a * a * a / mu)


class MeanMotion:
    r"""class used to calculate mean motion of an orbit

    .. note::

       mean motion will commonly be referenced as :math:`n` in documentation"""

    @staticmethod
    def from_a_mu(a: float, mu: float) -> float:
        r"""calculate the mean motion using equation 1-27 in :ref:`vallado`

        .. math::

           n = \frac{2\pi}{\tau} = \sqrt{\frac{\mu}{a^3}}

        :param a: semi-major axis in :math:`km`
        :type a: float
        :param mu: gravitational constant time mass of central body combined units of :math:`\frac{km^3}{s^2}`
        :type mu: float
        :return: mean motion in :math:`\frac{rad}{s}`
        :rtype: float
        """
        return sqrt(mu / (a * a * a))

    @staticmethod
    def from_tau(tau: float) -> float:
        r"""calculate mean motion using equation 1-27 in :ref:`vallado`

        .. math::

           n = \frac{2\pi}{\tau} = \sqrt{\frac{\mu}{a^3}}

        :param tau: period of orbit in :math:`s`
        :type tau: float
        :return: mean motion in :math:`\frac{rad}{s}`
        :rtype: float
        """
        return 2 * pi / tau


class EccentricAnomaly:
    r"""class used to solve eccentric anomaly

    .. note::

       eccentric anomaly will commonly be referenced as :math:`E` in documentation"""

    #: the tolerance used to stop the recursive solutions of Kepler's equation
    TOLERANCE: float = 1e-12

    @staticmethod
    def from_ma_e(ma: float, e: float) -> float:
        r"""calculate the eccentric anomaly using algorithm 2 in :ref:`vallado`

        .. math::

           E_{n+1} = E_n + \left(\frac{M-E_n+e\sin{(E_n)}}{1-e\cos{(E_n)}}\right)

        .. note::

           :math:`E_0 = M - e` for :math:`-\pi<M<0` or :math:`M>\pi`
           otherwise :math:`E_0 = M + e`

           looping continues until :math:`\vert E_{n+1} - E_n\vert < tolerance`

        :param ma: mean anomaly in :math:`rads`
        :type ma: float
        :param e: eccentricity
        :type e: float
        :return: eccentric anomaly in :math:`rads`
        :rtype: float
        """
        converged: bool = False
        ea0: float = ma

        if (ma > -pi and ma < 0) or ma > pi:
            ea0 -= e
        else:
            ea0 += e

        while not converged:
            ean = ea0 + (ma - ea0 + e * sin(ea0)) / (1 - e * cos(ea0))
            if abs(ean - ea0) < EccentricAnomaly.TOLERANCE:
                converged = True
            else:
                ea0 = ean

        if ean < 0:
            ean += 2 * pi
        return ean

    @staticmethod
    def from_rdv_r_a_n(r_dot_v: float, r: float, a: float, n: float) -> float:
        r"""calculate eccentric anomaly

        :param r_dot_v: dot product of position and velocity
        :type r_dot_v: float
        :param r: magnitude of position in :math:`km`
        :type r: float
        :param a: semi-major axis in :math:`km`
        :type a: float
        :param n: mean motion in :math:`\frac{rad}{s}`
        :type n: float
        :return: eccentric anomaly in :math:`rads`
        :rtype: float

        .. todo::

           find equation reference
        """
        ea: float = atan2(r_dot_v / (a * a * n), 1 - r / a)
        if ea < 0:
            ea += 2 * pi
        return ea


class TrueAnomaly:
    r"""static class used to solve true anomaly

    .. note::

       true anomaly will commonly be referenced as :math:`\nu` in documentation
    """

    @staticmethod
    def from_e_ea(e: float, ea: float) -> float:
        """calculate true anomaly

        :param e: eccentricity
        :type e: float
        :param ea: eccentric anomaly in :math:`rads`
        :type ea: float
        :return: true anomaly in :math:`rads`
        :rtype: float

        .. todo::

           find equation reference
        """
        ta: float = atan2(sqrt(1 - e * e) * sin(ea), cos(ea) - e)
        if ta < 0:
            ta += 2 * pi
        return ta


class ArgumentOfPerigee:
    r"""static class used to solve argument of perigee

    .. note::

       argument of perigee will commonly be referenced as :math:`\omega` in documentation
    """

    @staticmethod
    def from_u_nu(u: float, nu: float) -> float:
        r"""calculate the argument of perigee

        :param u: argument of latitude in :math:`rads`
        :type u: float
        :param nu: true anomaly in :math:`rads`
        :type nu: float
        :return: argument of perigee in :math:`rads`
        :rtype: float

        .. todo::

           find equation reference
        """
        w: float = u - nu
        if w < 0:
            w += 2 * pi
        return w


class ArgumentOfLatitude:
    """static class used to solve argument of latitude

    .. note::

       argument of latitude will commonly be referenced as :math:`u` in documentation
    """

    @staticmethod
    def from_r_w(r: Vector3D, w: Vector3D) -> float:
        r"""calculate the argument of latitude

        :param r: position vector in :math:`km`
        :type r: Vector3D
        :param w: normalized areal velocity vector in :math:`\frac{km^2}{s}`
        :type w: Vector3D
        :return: argument of latitude in :math:`rads`
        :rtype: float

        .. todo::

           find equation reference
        """
        u: float = atan2(r.z, -r.x * w.y + r.y * w.x)
        if u < 0:
            u += 2 * pi
        return u


class MeanAnomaly:
    """static class used to solve mean anomaly

    .. note::

       mean anomaly will commonly be referenced as :math:`M` in documentation
    """

    @staticmethod
    def from_ea_e(ea: float, e: float) -> float:
        r"""calculate mean anomaly using equation 2-4 in :ref:`vallado`

        .. math::

           M = E - e\sin{(E)}

        :param ea: eccentric anomaly in :math:`rads`
        :type ea: float
        :param e: eccentricity
        :type e: float
        :return: mean anomaly in :math:`rads`
        :rtype: float
        """
        ma: float = ea - e * sin(ea)
        if ma < 0:
            ma += 2 * pi
        return ma


class EquationsOfMotion:
    """class used to solve equations of motion"""

    #: used to solve semi-major axis :math:`a`
    A = SemiMajorAxis

    #: used to solve semi-minor axis :math:`b`
    B = SemiMinorAxis

    #: used to solve semi-parameter :math:`p`
    P = SemiParameter

    #: used to solve eccentricity :math:`e`
    E = Eccentricity

    #: used to solve period :math:`tau`
    TAU = Period

    #: used to solve mean motion :math:`n`
    N = MeanMotion

    #: used to solve velocity :math:`v`
    V = VisVivaVelocity

    #: used to solve specific mechanical energy :math:`\xi`
    XI = SpecificMechanicalEnergy

    #: used to solve areal velocity :math:`h`
    H = ArealVelocity

    #: used to solve flattening :math:`f`
    F = Flattening

    #: used to solve eccentric anomaly :math:`E`
    EA = EccentricAnomaly

    #: used to solve inclination :math:`i`
    I = Inclination

    #: used to solve raan :math:`\Omega`
    RAAN = RightAscensionNode

    #: used to solve true anomaly :math:`\nu`
    NU = TrueAnomaly

    #: used to solve argument of perigee :math:`\omega`
    W = ArgumentOfPerigee

    #: used to solve argument of latitude :math:`u`
    U = ArgumentOfLatitude

    #: used to solve mean anomaly :math:`M`
    MA = MeanAnomaly



import unittest

from pysmad.bodies.celestial import Earth, Moon, Sun
from pysmad.math.linalg import Vector3D
from pysmad.time import Epoch


class TestEarth(unittest.TestCase):
    def test_mu(self):
        self.assertEqual(398600.4418, Earth.MU)

    def test_radius(self):
        self.assertEqual(6378.137, Earth.RADIUS)

    def test_flattening(self):
        self.assertEqual(1 / 298.2572235, Earth.FLATTENING)

    def test_c(self):
        c00 = Earth.C[0][0]
        c10 = Earth.C[1][0]
        c11 = Earth.C[1][1]
        c20 = Earth.C[2][0]
        c21 = Earth.C[2][1]
        c22 = Earth.C[2][2]
        c30 = Earth.C[3][0]
        c31 = Earth.C[3][1]
        c32 = Earth.C[3][2]
        c33 = Earth.C[3][3]
        c40 = Earth.C[4][0]
        c41 = Earth.C[4][1]
        c42 = Earth.C[4][2]
        c43 = Earth.C[4][3]
        c44 = Earth.C[4][4]
        self.assertAlmostEqual(1, c00, 15)
        self.assertAlmostEqual(0, c10, 15)
        self.assertAlmostEqual(0, c11, 15)
        self.assertAlmostEqual(-0.0010826261738522227, c20, 15)
        self.assertAlmostEqual(-2.6673947523748365e-10, c21, 15)
        self.assertAlmostEqual(1.574615325722917e-06, c22, 15)
        self.assertAlmostEqual(2.5324105185677225e-06, c30, 15)
        self.assertAlmostEqual(2.1931496313133285e-06, c31, 15)
        self.assertAlmostEqual(3.0904390039164885e-07, c32, 15)
        self.assertAlmostEqual(1.0058351340882285e-07, c33, 15)
        self.assertAlmostEqual(1.6198975999169731e-06, c40, 15)
        self.assertAlmostEqual(-5.086435604395839e-07, c41, 15)
        self.assertAlmostEqual(7.837454574045524e-08, c42, 15)
        self.assertAlmostEqual(5.9215017763966597e-08, c43, 15)
        self.assertAlmostEqual(-3.983204248731872e-09, c44, 15)


class TestSun(unittest.TestCase):

    EPOCH: Epoch = Epoch.from_gregorian(2022, 2, 25, 0, 1, 9.184)
    TRUTH: Vector3D = Vector3D(1.353158384133262e8, -5.514968448042840e7, -2.390803633125914e7)

    def test_get_position(self):
        sun_pos = Sun.get_position(self.EPOCH)
        self.assertAlmostEqual(sun_pos.magnitude() / self.TRUTH.magnitude(), 1, 4)
        self.assertAlmostEqual(sun_pos.angle(self.TRUTH), 0, 2)


class TestMoon(unittest.TestCase):

    EPOCH: Epoch = Epoch.from_gregorian(2022, 2, 25, 0, 1, 9.184)
    TRUTH: Vector3D = Vector3D(-6.454159844478600e4, -3.280761448809440e5, -1.566863311585961e5)

    def test_get_position(self):
        moon_pos = Moon.get_position(self.EPOCH)
        self.assertAlmostEqual(moon_pos.magnitude() / self.TRUTH.magnitude(), 1, 4)
        self.assertAlmostEqual(moon_pos.angle(self.TRUTH), 0, 3)



from math import asin, atan2, cos, pi, sin, sqrt

from pysmad.bodies.celestial import Earth
from pysmad.math.functions import sign
from pysmad.math.linalg import Matrix3D, Vector3D
from pysmad.time import Epoch


class SphericalPosition:
    def __init__(self, r: float, ra: float, dec: float) -> None:
        """class used to perform spherical transformations

        :param r: magnitude of the vector
        :type r: float
        :param ra: right ascension of the vector (radians)
        :type ra: float
        :param dec: declination of the vector (radians)
        :type dec: float
        """
        #: magnitude of the vector
        self.radius: float = r

        #: right ascension of the vector in radians
        self.right_ascension: float = ra

        #: declination of the vector in radians
        self.declination: float = dec

    @classmethod
    def from_cartesian(cls, pos: Vector3D) -> "SphericalPosition":
        """create a spherical vector using cartesian components

        :param pos: cartesian vector
        :type pos: Vector3D
        :return: position represented with spherical components
        :rtype: SphericalPosition
        """
        ra: float = atan2(pos.y, pos.x)
        if ra < 0:
            ra += 2 * pi
        dec: float = atan2(pos.z, sqrt(pos.x * pos.x + pos.y * pos.y))
        return cls(pos.magnitude(), ra, dec)

    def to_cartesian(self) -> Vector3D:
        """calculate the vector as represented by x, y, and z

        :return: vector of equal magnitude in direction but in cartesian coordinates
        :rtype: Vector3D
        """
        cd: float = cos(self.declination)
        return Vector3D(cd * cos(self.right_ascension), cd * sin(self.right_ascension), sin(self.declination)).scaled(
            self.radius
        )


class ENZ:
    @staticmethod
    def matrix(lamb: float, psi: float) -> Matrix3D:
        """calculate the transformation matrix required for ENZ/ITRF transformations

        :param lamb: geodetic longitude
        :type lamb: float
        :param psi: geodetic latitude
        :type psi: float
        :return: matrix used to go from ITRF to ENZ
        :rtype: Matrix3D
        """
        slamb: float = sin(lamb)
        clamb: float = cos(lamb)
        spsi: float = sin(psi)
        cpsi: float = cos(psi)

        return Matrix3D(
            Vector3D(-slamb, clamb, 0.0),
            Vector3D(-spsi * clamb, -spsi * slamb, cpsi),
            Vector3D(cpsi * clamb, cpsi * slamb, spsi),
        )


class LLA:
    def __init__(self, lat: float, longit: float, alt: float) -> None:
        """used to perform operations for a state in an oblate earth frame

        :param lat: geodetic latitude in radians
        :type lat: float
        :param long: geodetic longitude in radians
        :type long: float
        :param alt: altitude above the surface in km
        :type alt: float
        """
        #: geodetic latitude in radians
        self.latitude: float = lat

        #: geodetic longitude in radians
        self.longitude: float = longit

        #: altitude above the surface in km
        self.altitude: float = alt

    def copy(self) -> "LLA":
        """creates a duplicate of the calling state

        :return: state with properties that match that of the calling state
        :rtype: LLA
        """
        return LLA(self.latitude, self.longitude, self.altitude)


class _PositionConvertGCRF:
    """class used to convert GCRF positions to other frames"""

    @staticmethod
    def to_itrf(pos: Vector3D, epoch: Epoch) -> Vector3D:
        """calculate the ITRF position

        :param pos: GCRF position
        :type pos: Vector3D
        :param epoch: epoch for which the position is valid
        :type epoch: Epoch
        :return: ITRF position
        :rtype: Vector3D
        """
        return Earth.rotation(epoch).multiply_vector(_PositionConvertGCRF.to_tod(pos, epoch))

    @staticmethod
    def to_tod(pos: Vector3D, epoch: Epoch) -> Vector3D:
        """calculate the TOD position

        :param pos: GCRF position
        :type pos: Vector3D
        :param epoch: epoch for which the position is valid
        :type epoch: Epoch
        :return: TOD position
        :rtype: Vector3D
        """
        return Earth.nutation(epoch).multiply_vector(_PositionConvertGCRF.to_mod(pos, epoch))

    @staticmethod
    def to_mod(pos: Vector3D, epoch: Epoch) -> Vector3D:
        """calculate the MOD position

        :param pos: GCRF position
        :type pos: Vector3D
        :param epoch: epoch for which the position is valid
        :type epoch: Epoch
        :return: MOD position
        :rtype: Vector3D
        """
        return Earth.precession(epoch).multiply_vector(pos)

    @staticmethod
    def to_ijk(pos: Vector3D, epoch: Epoch) -> Vector3D:
        """calculate the IJK position

        :param pos: GCRF position
        :type pos: Vector3D
        :param epoch: epoch for which the position is valid
        :type epoch: Epoch
        :return: IJK position
        :rtype: Vector3D
        """
        return _PositionConvertITRF.to_ijk(_PositionConvertGCRF.to_itrf(pos, epoch), epoch)


class _PositionConvertITRF:
    """class used to convert ITRF positions to other frames"""

    @staticmethod
    def to_gcrf(pos: Vector3D, epoch: Epoch) -> Vector3D:
        """calculate the GCRF position

        :param pos: ITRF position
        :type pos: Vector3D
        :param epoch: epoch for which the position is valid
        :type epoch: Epoch
        :return: GCRF position
        :rtype: Vector3D
        """
        return Earth.precession(epoch).transpose().multiply_vector(_PositionConvertITRF.to_mod(pos, epoch))

    @staticmethod
    def to_tod(pos: Vector3D, epoch: Epoch) -> Vector3D:
        """calculate the TOD position

        :param pos: ITRF position
        :type pos: Vector3D
        :param epoch: epoch for which the position is valid
        :type epoch: Epoch
        :return: TOD position
        :rtype: Vector3D
        """
        return Earth.rotation(epoch).transpose().multiply_vector(pos)

    @staticmethod
    def to_mod(pos: Vector3D, epoch: Epoch) -> Vector3D:
        """calculate the MOD position

        :param pos: ITRF position
        :type pos: Vector3D
        :param epoch: epoch for which the position is valid
        :type epoch: Epoch
        :return: MOD position
        :rtype: Vector3D
        """
        return Earth.nutation(epoch).transpose().multiply_vector(_PositionConvertITRF.to_tod(pos, epoch))

    @staticmethod
    def to_lla(pos: Vector3D) -> LLA:
        """calculate the LLA position

        :param pos: ITRF position
        :type pos: Vector3D
        :return: LLA position
        :rtype: LLA
        """
        x: float = pos.x
        y: float = pos.y
        z: float = pos.z

        # Equation 2.77a
        a: float = Earth.RADIUS
        a2: float = a * a
        f: float = Earth.FLATTENING
        b: float = a - f * a
        b2: float = b * b
        e2: float = 1 - b2 / a2
        eps2: float = a2 / b2 - 1.0
        rho: float = sqrt(x * x + y * y)

        # Equation 2.77b
        p: float = abs(z) / eps2
        s: float = rho * rho / (e2 * eps2)
        q: float = p * p - b2 + s

        # Equation 2.77c
        u: float = p / sqrt(q)
        v: float = b2 * u * u / q
        cap_p: float = 27.0 * v * s / q
        cap_q: float = (sqrt(cap_p + 1) + sqrt(cap_p)) ** (2.0 / 3.0)

        # Equation 2.77d
        t: float = (1.0 + cap_q + 1.0 / cap_q) / 6.0
        c: float = sqrt(u * u - 1.0 + 2.0 * t)
        w: float = (c - u) / 2.0

        # Equation 2.77e
        base: float = sqrt(t * t + v) - u * w - t / 2.0 - 0.25
        if base < 0:
            base = 0
        arg: float = w + sqrt(base)
        d: float = sign(z) * sqrt(q) * arg

        # Equation 2.77f
        n: float = a * sqrt(1.0 + eps2 * d * d / b2)
        arg = (eps2 + 1.0) * (d / n)
        lamb: float = asin(arg)

        # Equation 2.77g
        h: float = rho * cos(lamb) + z * sin(lamb) - a2 / n
        phi: float = atan2(y, x)
        if phi < 0:
            phi += pi * 2.0

        return LLA(lamb, phi, h)

    @staticmethod
    def to_ijk(pos: Vector3D, epoch: Epoch) -> Vector3D:
        """calculate the IJK position

        :param pos: ITRF position
        :type pos: Vector3D
        :param epoch: epoch for which the position is valid
        :type epoch: Epoch
        :return: IJK position
        :rtype: Vector3D
        """
        return pos.rotation_about_axis(Vector3D(0, 0, 1), epoch.greenwich_hour_angle())


class _PositionConvertLLA:
    """class used to convert LLA positions to other frames"""

    @staticmethod
    def to_itrf(lla: LLA) -> Vector3D:
        """calculate the ITRF position

        :param lla: LLA position
        :type lla: LLA
        :return: ITRF position
        :rtype: Vector3D
        """
        lat: float = lla.latitude
        longitude: float = lla.longitude
        alt: float = lla.altitude

        f: float = Earth.FLATTENING
        e: float = sqrt(f * (2 - f))
        slat: float = sin(lat)
        clat: float = cos(lat)
        n: float = Earth.RADIUS / sqrt(1 - e * e * slat * slat)

        return Vector3D(
            (n + alt) * clat * cos(longitude), (n + alt) * clat * sin(longitude), (n * (1.0 - e * e) + alt) * slat
        )


class _PositionConvertENZ:
    """class used to convert ENZ positions to other frames"""

    @staticmethod
    def to_itrf(lla: LLA, enz: Vector3D) -> Vector3D:
        """calculate the ITRF position

        :param lla: LLA position of the ENZ origin
        :type lla: LLA
        :param enz: ENZ position from the lla origin
        :type enz: Vector3D
        :return: ITRF position
        :rtype: Vector3D
        """
        return ENZ.matrix(lla.longitude, lla.latitude).transpose().multiply_vector(enz)


class PositionConvert:

    #: used to perform conversions from GCRF
    gcrf = _PositionConvertGCRF

    #: used to perform conversions from ITRF
    itrf = _PositionConvertITRF

    #: used to perform conversions from LLA
    lla = _PositionConvertLLA

    #: used to perform conversions from ENZ
    enz = _PositionConvertENZ



import unittest
from math import radians

from pysmad.math.linalg import Vector3D


class TestVector3D(unittest.TestCase):

    VEC_1 = Vector3D(4, 2, 42)
    VEC_2 = Vector3D(4, 42, 2)
    ROT_VEC = Vector3D(4, 2, 4)
    ROT_AXIS = Vector3D(6, 8, 2)
    ANGLE: float = radians(25)

    def test_rotation_about_axis(self):
        vec = self.ROT_VEC.rotation_about_axis(self.ROT_AXIS, self.ANGLE)
        self.assertAlmostEqual(vec.x, 5.045038149)
        self.assertAlmostEqual(vec.y, 1.4954979717)
        self.assertAlmostEqual(vec.z, 2.8828936665)

    def test_plus(self):
        vec_sum = self.VEC_1.plus(self.VEC_2)
        self.assertAlmostEqual(vec_sum.x, 8)
        self.assertAlmostEqual(vec_sum.y, 44)
        self.assertAlmostEqual(vec_sum.z, 44)

    def test_minus(self):
        vec_diff = self.VEC_1.minus(self.VEC_2)
        self.assertAlmostEqual(vec_diff.x, 0)
        self.assertAlmostEqual(vec_diff.y, -40)
        self.assertAlmostEqual(vec_diff.z, 40)

    def test_dot(self):
        vec_dot = self.VEC_1.dot(self.VEC_2)
        self.assertAlmostEqual(vec_dot, 184)

    def test_cross(self):
        vec_cross = self.VEC_1.cross(self.VEC_2)
        self.assertAlmostEqual(vec_cross.x, -1760)
        self.assertAlmostEqual(vec_cross.y, 160)
        self.assertAlmostEqual(vec_cross.z, 160)

    def test_magnitude(self):
        self.assertAlmostEqual(self.VEC_1.magnitude(), 42.23742416388575)
        self.assertAlmostEqual(self.VEC_2.magnitude(), 42.23742416388575)

    def test_scaled(self):
        vec_scaled = self.VEC_1.scaled(2)
        self.assertAlmostEqual(vec_scaled.x, 8)
        self.assertAlmostEqual(vec_scaled.y, 4)
        self.assertAlmostEqual(vec_scaled.z, 84)

    def test_normalized(self):
        vec_normed = self.VEC_1.normalized()
        self.assertAlmostEqual(vec_normed.x, 0.0947027447620757)
        self.assertAlmostEqual(vec_normed.y, 0.0473513723810378)
        self.assertAlmostEqual(vec_normed.z, 0.9943788200017946)



import unittest
from math import degrees

from pysmad.time import Epoch


class TestEpoch(unittest.TestCase):

    EPOCH = Epoch.from_gregorian(2022, 12, 19, 12, 1, 9.184)

    def test_julian_value(self):
        self.assertAlmostEqual(self.EPOCH.julian_value(), 2459933.00080074)

    def test_greenwich_hour_angle(self):
        self.assertAlmostEqual(degrees(self.EPOCH.greenwich_hour_angle()), 268.0707472726, 5)



from math import sqrt

from pysmad.bodies.celestial import Earth


class VisViva:
    """class used to perform calculations related to the vis-viva law

    :return: velocities for various orbit conditions
    :rtype: float
    """

    @staticmethod
    def circular_velocity(r: float) -> float:
        """calculate the velocity of a circular orbit

        :param r: sma of the circular orbit in km
        :type r: float
        :return: velocity in km/s
        :rtype: float
        """
        return sqrt(Earth.MU / r)

    @staticmethod
    def eccentric_velocity(r: float, a: float) -> float:
        """calculate the velocity at a position in an eccentric orbit

        :param r: current radius in km
        :type r: float
        :param a: semi-major axis in km
        :type a: float
        :return: velocity at the given radius in km/s
        :rtype: float
        """
        return sqrt(Earth.MU * (2 / r - 1 / a))


class Hohmann:
    def __init__(self, r1: float, r2: float) -> None:
        """calculate details of a hohmann transfer going from one circular orbit to another

        :param r1: sma of initial circular orbit
        :type r1: float
        :param r2: sma of final circular orbit
        :type r2: float
        """
        sma: float = (r1 + r2) / 2
        #: velocity prior to the first burn in km/s
        self.initial_velocity: float = VisViva.circular_velocity(r1)

        #: velocity after the first burn in km/s
        self.transfer_velocity_1: float = VisViva.eccentric_velocity(r1, sma)

        #: velocity prior to the second burn in km/s
        self.transfer_velocity_2: float = VisViva.eccentric_velocity(r2, sma)

        #: velocity after the second burn in km/s
        self.final_velocity: float = VisViva.circular_velocity(r2)

        #: magnitude of the first burn in km/s
        self.delta_v_1: float = self.transfer_velocity_1 - self.initial_velocity

        #: magnitude of the second burn in km/s
        self.delta_v_2: float = self.final_velocity - self.transfer_velocity_2



import matplotlib.pyplot as plt

from pysmad.bodies.artificial import Spacecraft
from pysmad.coordinates.states import GCRF, HCW, StateConvert
from pysmad.math.linalg import Vector3D, Vector6D
from pysmad.time import Epoch

# Create scenario start epoch
start_epoch: Epoch = Epoch.from_gregorian(2022, 12, 20, 0, 0, 0)

# Create ECI state for target
target_state: GCRF = GCRF(start_epoch, Vector3D(42164, 0, 0), Vector3D(0, 3.075, 0))
target: Spacecraft = Spacecraft(target_state)

# Step target forward 6 hours
target.step_to_epoch(start_epoch.plus_days(0.25))

# Define propagation end of 1 day after current target epoch
end_epoch = target.current_epoch().plus_days(1)

# Create chase state relative to target
rel_chase_state = HCW.from_state_vector(Vector6D(-11, 0, 0, 0, 0.0016, 0))
chase_state: GCRF = StateConvert.hcw.to_gcrf(rel_chase_state, target.current_state())
chase: Spacecraft = Spacecraft(chase_state)

# Create empty data lists
times, detections, resolutions = [], [], []

# Place chase in target-tracking attitude
chase.track_state(target)

# Propagate
while chase.current_epoch().value < end_epoch.value:

    # Step vehicles and sync epochs
    chase.step()
    target.step_to_epoch(chase.current_epoch())

    # Store epochs
    times.append(chase.current_epoch().value)

    # Store resolutions
    resolutions.append(chase.nfov.resolution(chase.range(target)) * 1e5)

    # Use binary reference of visible (1) or not (0)
    if chase.detect(target):
        detections.append(1)
    else:
        detections.append(0)

# Plot data
plt.plot(times, detections)
plt.plot(times, resolutions)
plt.show()



import unittest
from math import radians

from pysmad.bodies.terrestrial import GroundSite
from pysmad.coordinates.elements import ClassicalElements
from pysmad.coordinates.states import ITRF
from pysmad.estimation.iod import Gauss
from pysmad.estimation.obs import GroundObservation
from pysmad.math.linalg import Vector3D
from pysmad.time import Epoch


class TestGauss(unittest.TestCase):
    SITE: GroundSite = GroundSite.from_itrf_position(Vector3D(1344.143, 6068.601, 1429.311))

    def test_coes_from_positions(self):
        state_1: ITRF = ITRF(Epoch.from_gregorian(1999, 4, 2, 0, 31, 9.184), self.SITE.itrf_position, Vector3D(0, 0, 0))
        ob_1: GroundObservation = GroundObservation.from_angles_and_range(
            state_1, radians(132.67), radians(32.44), 16945.45, 0, 0
        )
        state_2: ITRF = ITRF(Epoch.from_gregorian(1999, 4, 2, 3, 1, 9.184), self.SITE.itrf_position, Vector3D(0, 0, 0))
        ob_2: GroundObservation = GroundObservation.from_angles_and_range(
            state_2, radians(123.08), radians(50.06), 37350.34, 0, 0
        )

        coes: ClassicalElements = Gauss.coes_from_positions(ob_1, ob_2)

        self.assertAlmostEqual(coes.semimajor_axis, 28196.77578363615)
        self.assertAlmostEqual(coes.eccentricity, 0.7679436)
        self.assertAlmostEqual(coes.inclination, 0.3545688111634284)
        self.assertAlmostEqual(coes.raan, 6.268254918583486)
        self.assertAlmostEqual(coes.argument_of_perigee, 3.1315625245621277)
        self.assertAlmostEqual(coes.mean_anomaly, 0.5102678870735644)

    def test_site_itrf_position(self):
        self.assertAlmostEqual(self.SITE.itrf_position.x, 1344.143)
        self.assertAlmostEqual(self.SITE.itrf_position.y, 6068.601)
        self.assertAlmostEqual(self.SITE.itrf_position.z, 1429.311)

    def test_site_eci_position(self):
        gmst: float = Epoch.from_gregorian(1999, 4, 2, 0, 31, 9.184).greenwich_hour_angle()
        eci: Vector3D = self.SITE.itrf_position.rotation_about_axis(Vector3D(0, 0, 1), gmst)
        self.assertAlmostEqual(eci.x, 534.388240046999)
        self.assertAlmostEqual(eci.y, -6192.662408903832)
        self.assertAlmostEqual(eci.z, 1429.3110)



import os
from math import asin, atan2, cos, sin, sqrt, tan
from typing import List

from pysmad.bodies.celestial import Earth, Moon, Sun
from pysmad.coordinates.positions import PositionConvert, SphericalPosition
from pysmad.math.constants import BASE_IN_KILO
from pysmad.math.functions import LegendrePolynomial
from pysmad.math.linalg import Matrix3D, Vector3D, Vector6D
from pysmad.time import Epoch


class State:
    def __init__(self, epoch: Epoch, r: Vector3D, v: Vector3D) -> None:
        """class used to perform operations for time-dependent states

        :param epoch: time for which the position and velocity are valid
        :type epoch: Epoch
        :param r: position of the state
        :type r: Vector3D
        :param v: velocity of the state
        :type v: Vector3D
        """
        #: time for which the position and velocity are valid
        self.epoch: Epoch = epoch.copy()

        #: position of the state
        self.position: Vector3D = r.copy()

        #: velocity of the state
        self.velocity: Vector3D = v.copy()

        #: state vector whose elements are equal to that of the position and velocity unpacked
        self.vector = Vector6D.from_position_and_velocity(self.position, self.velocity)

    def copy(self):
        """create a duplicate of the calling state"""
        return type(self)(self.epoch, self.position, self.velocity)

    def vector_list(self) -> List[Vector3D]:
        """create a list of the position and velocity vectors

        :return: list of indices 0 == position and 1 == velocity
        :rtype: List[Vector3D]
        """
        return [self.position, self.velocity]

    @staticmethod
    def csv_headers():
        return ",".join(["TARGET_ID", "UTC_EPOCH", "FRAME", "X", "Y", "Z", "VX", "VY", "VZ"])


class LiveVector(State):
    def __init__(self, vec_dict: dict) -> None:
        self.epoch: Epoch = Epoch.from_udl_string(vec_dict["epoch"])
        self.position: Vector3D = Vector3D(vec_dict["xpos"], vec_dict["ypos"], vec_dict["zpos"])
        self.velocity = Vector3D(vec_dict["xvel"], vec_dict["yvel"], vec_dict["zvel"])
        self.vector = Vector6D.from_position_and_velocity(self.position, self.velocity)
        self.reference_frame: str = vec_dict.get("referenceFrame", "J2000")
        self.srp_coefficient: float = vec_dict.get("solarRadPressCoeff", None)
        self.covariance: list[float] = vec_dict.get("cov", None)
        self.sat_id: str = vec_dict.get("origObjectId", vec_dict.get("idOnOrbit", "UNKNOWN"))

    def get_gcrf_state(self) -> "GCRF":
        return GCRF(self.epoch, self.position, self.velocity)

    def to_csv_format(self) -> str:
        return ",".join(
            [
                self.sat_id,
                self.epoch.to_udl_string(),
                self.reference_frame,
                str(self.position.x),
                str(self.position.y),
                str(self.position.z),
                str(self.velocity.x),
                str(self.velocity.y),
                str(self.velocity.z),
            ]
        )


class LiveVectorSet:
    def __init__(self):
        self.list = []
        self.sensors = {}
        self.targets = {}
        self.sources = {}
        self.modes = {}
        self.total = 0

    def publish_set(self, filename: str, write_mode: str) -> None:

        write_header = False
        if not os.path.exists(filename):
            write_header = True

        with open(filename, write_mode) as f:
            if write_header:
                f.write(State.csv_headers() + "\n")
            for state in self.list:
                f.write("".join([state.to_csv_format(), "\n"]))

    def process_vector(self, state: LiveVector) -> None:
        self.list.append(state)
        self.total += 1
        if self.targets.get(state.sat_id) is None:
            self.targets[state.sat_id] = state
        elif self.targets[state.sat_id].epoch.value < state.epoch.value:
            self.targets[state.sat_id] = state

    def get_latest(self, scc: str) -> LiveVector:
        return self.targets[scc].get_gcrf_state()


class ITRF(State):
    def __init__(self, epoch: Epoch, r: Vector3D, v: Vector3D) -> None:
        """class used to perform operations and modeling in the International Terrestrial Reference Frame

        :param epoch: time for which the position and velocity are valid
        :type epoch: Epoch
        :param r: position of the state
        :type r: Vector3D
        :param v: velocity of the state
        :type v: Vector3D
        """
        super().__init__(epoch, r, v)

    @classmethod
    def from_fixed(cls, epoch: Epoch, r: Vector3D) -> "ITRF":
        """create an ITRF state from a position fixed to the Earth surface

        :param epoch: time for which the position and velocity are valid
        :type epoch: Epoch
        :param r: position of the state
        :type r: Vector3D
        :return: ITRF state with velocity vector == 0
        :rtype: ITRF
        """
        return cls(epoch, r, Vector3D(0, 0, 0))


class HCW(State):
    def __init__(self, epoch: Epoch, r: Vector3D, v: Vector3D) -> None:
        """class used to perform operations and modeling in the HCW Frame

        :param epoch: time for which the position and velocity are valid
        :type epoch: Epoch
        :param r: position of the state
        :type r: Vector3D
        :param v: velocity of the state
        :type v: Vector3D
        """
        super().__init__(epoch, r, v)

    @classmethod
    def from_state_vector(cls, state_vec: Vector6D) -> "HCW":
        """create a hill state of from a full state vector

        :param state_vec: 6-dimensional vector of position and velocity
        :type state_vec: Vector6D
        :return: HCW state with arbitrary epoch
        :rtype: HCW
        """
        return cls(
            Epoch(0),
            Vector3D(state_vec.x, state_vec.y, state_vec.z),
            Vector3D(state_vec.vx, state_vec.vy, state_vec.vz),
        )

    @staticmethod
    def frame_matrix(origin: "GCRF") -> Matrix3D:
        """create a radial, in-track, cross-track axes matrix

        :param origin: inertial state that acts as the origin for the RIC frame
        :type origin: GCRF
        :return: matrix with rows of radial, in-track, and cross-track
        :rtype: Matrix3D
        """
        r: Vector3D = origin.position.normalized()
        c: Vector3D = origin.position.cross(origin.velocity).normalized()
        i: Vector3D = c.cross(r)
        return Matrix3D(r, i, c)


class GCRF(State):
    def __init__(self, epoch: Epoch, r: Vector3D, v: Vector3D) -> None:
        """class used to perform operations and modeling in the Geocentric Celestial Reference Frame

        :param epoch: time for which the position and velocity are valid
        :type epoch: Epoch
        :param r: position of the state
        :type r: Vector3D
        :param v: velocity of the state
        :type v: Vector3D
        """
        super().__init__(epoch, r, v)

        #: acceleration due to thrust
        self.thrust: Vector3D = Vector3D(0, 0, 0)

        #: scalar used for srp acceleration calculations
        self.srp_scalar: float = 0

        #: boolean to determine if perturbations are modeled during propagation
        self.use_perturbations: bool = True

    def acceleration_from_gravity(self) -> Vector3D:
        """calculates the gravity due to a nonspherical earth

        :return: vector representing the acceleration due to gravity
        :rtype: Vector3D
        """
        ecef: Vector3D = PositionConvert.gcrf.to_itrf(self.position, self.epoch)
        sphr_pos: SphericalPosition = SphericalPosition.from_cartesian(ecef)
        p: List[List[float]] = LegendrePolynomial(sphr_pos.declination).p

        m: int = 0
        n: int = 2

        partial_r: float = 0
        partial_phi: float = 0
        partial_lamb: float = 0
        recip_r: float = 1 / self.position.magnitude()
        mu_over_r: float = Earth.MU * recip_r
        r_over_r: float = Earth.RADIUS * recip_r
        r_exponent: float = 0
        clam: float = 0
        slam: float = 0
        recip_root: float = 1 / sqrt(ecef.x * ecef.x + ecef.y * ecef.y)
        rz_over_root: float = ecef.z * recip_r * recip_r * recip_root
        while n < Earth.DEGREE_AND_ORDER:
            m = 0
            r_exponent = r_over_r**n
            while m <= n:
                clam = cos(m * sphr_pos.right_ascension)
                slam = sin(m * sphr_pos.right_ascension)
                partial_r += r_exponent * (n + 1) * p[n][m] * (Earth.C[n][m] * clam + Earth.S[n][m] * slam)
                partial_phi += (
                    r_exponent
                    * (p[n][m + 1] - m * tan(sphr_pos.declination) * p[n][m])
                    * (Earth.C[n][m] * clam + Earth.S[n][m] * slam)
                )
                partial_lamb += r_exponent * m * p[n][m] * (Earth.S[n][m] * clam - Earth.C[n][m] * slam)

                m += 1

            n += 1

        partial_r *= -recip_r * mu_over_r
        partial_phi *= mu_over_r
        partial_lamb *= mu_over_r

        return PositionConvert.itrf.to_gcrf(
            Vector3D(
                (recip_r * partial_r - rz_over_root * partial_phi) * ecef.x
                - (recip_root * recip_root * partial_lamb) * ecef.y,
                (recip_r * partial_r - rz_over_root * partial_phi) * ecef.y
                + (recip_root * recip_root * partial_lamb) * ecef.x,
                recip_r * partial_r * ecef.z + (1 / recip_root) * recip_r * recip_r * partial_phi,
            ),
            self.epoch,
        )

    def acceleration_from_earth(self) -> Vector3D:
        """calculate the acceleration on the state due to earth's gravity

        :return: vector representing the acceleration from earth
        :rtype: Vector3D
        """
        r_mag: float = self.position.magnitude()

        return self.position.scaled(-Earth.MU / (r_mag * r_mag * r_mag))

    def acceleration_from_moon(self) -> Vector3D:
        """calculate the acceleration on the state due to the moon

        :return: vector representing the acceleration from the moon
        :rtype: Vector3D
        """
        s: Vector3D = Moon.get_position(self.epoch)
        r: Vector3D = s.minus(self.position)
        r_mag: float = r.magnitude()
        s_mag: float = s.magnitude()
        vec_1: Vector3D = r.scaled(1 / (r_mag * r_mag * r_mag))
        vec_2: Vector3D = s.scaled(1 / (s_mag * s_mag * s_mag))
        return vec_1.minus(vec_2).scaled(Moon.MU)

    def acceleration_from_sun(self) -> Vector3D:
        """calculate the acceleration on the state due to the sun

        :return: vector representing the acceleration from the sun
        :rtype: Vector3D
        """
        s: Vector3D = Sun.get_position(self.epoch)
        r: Vector3D = s.minus(self.position)
        r_mag: float = r.magnitude()
        s_mag: float = s.magnitude()
        vec_1: Vector3D = r.scaled(1 / (r_mag * r_mag * r_mag))
        vec_2: Vector3D = s.scaled(1 / (s_mag * s_mag * s_mag))
        return vec_1.minus(vec_2).scaled(Sun.MU)

    def acceleration_from_srp(self) -> Vector3D:
        """calculate the acceleration on the state from solar radiation pressure

        :return: vector representing the acceleration from srp
        :rtype: Vector3D
        """
        sun_vec: Vector3D = self.sun_vector().normalized()
        s_mag: float = sun_vec.magnitude()
        return sun_vec.scaled(-Sun.P * self.srp_scalar / (s_mag * s_mag * BASE_IN_KILO))

    def acceleration_from_thrust(self) -> Vector3D:
        """retrieve the stored acceleration to be applied from thrusters

        :return: the current acceleration vector in the GCRF frame
        :rtype: Vector3D
        """
        return self.thrust.copy()

    def derivative(self) -> List[Vector3D]:
        """create a list with elements 0 == velocity and 1 == acceleration

        :return: list of velocity and acceleration
        :rtype: List[Vector3D]
        """
        net_a: Vector3D = self.acceleration_from_thrust()
        net_a = net_a.plus(self.acceleration_from_earth())
        if self.use_perturbations:
            net_a = net_a.plus(self.acceleration_from_moon())
            net_a = net_a.plus(self.acceleration_from_sun())
            net_a = net_a.plus(self.acceleration_from_srp())
            net_a = net_a.plus(self.acceleration_from_gravity())
        return [self.velocity.copy(), net_a]

    def sun_vector(self) -> Vector3D:
        """create a vector pointing from the calling state to the sun

        :return: vector originating at the calling state and terminating at the sun
        :rtype: Vector3D
        """
        return Sun.get_position(self.epoch).minus(self.position)

    def moon_vector(self) -> Vector3D:
        """create a vector pointing from the calling state to the moon

        :return: vector originating at the calling state and terminating at the moon
        :rtype: Vector3D
        """
        return Moon.get_position(self.epoch).minus(self.position)


class IJK(State):
    def __init__(self, epoch: Epoch, r: Vector3D, v: Vector3D) -> None:
        """class used to perform operations and modeling in the Geocentric Equatorial Coordinate System

        :param epoch: time for which the position and velocity are valid
        :type epoch: Epoch
        :param r: position of the state
        :type r: Vector3D
        :param v: velocity of the state
        :type v: Vector3D
        """
        super().__init__(epoch, r, v)


class _StateConvertGCRF:
    """class used to perform state conversions from GCRF"""

    @staticmethod
    def to_hcw(origin: GCRF, state: GCRF) -> HCW:
        """create a state in the HCW frame

        :param origin: state which represents the origin of the relative frame
        :type origin: GCRF
        :param state: state to be modeled in the Hill frame
        :type state: GCRF
        :return: HCW state
        :rtype: HCW
        """
        magrtgt: float = origin.position.magnitude()
        magrint: float = state.position.magnitude()
        rot_eci_rsw: Matrix3D = HCW.frame_matrix(origin)
        vtgtrsw: Vector3D = rot_eci_rsw.multiply_vector(origin.velocity)
        rintrsw: Vector3D = rot_eci_rsw.multiply_vector(state.position)
        vintrsw: Vector3D = rot_eci_rsw.multiply_vector(state.velocity)

        sinphiint: float = rintrsw.z / magrint
        phiint: float = asin(sinphiint)
        cosphiint: float = cos(phiint)
        lambdaint: float = atan2(rintrsw.y, rintrsw.x)
        sinlambdaint: float = sin(lambdaint)
        coslambdaint: float = cos(lambdaint)
        lambdadottgt: float = vtgtrsw.y / magrtgt

        r_hcw: Vector3D = Vector3D(magrint - magrtgt, lambdaint * magrtgt, phiint * magrtgt)

        rot_rsw_sez: Matrix3D = Matrix3D(
            Vector3D(sinphiint * coslambdaint, sinphiint * sinlambdaint, -cosphiint),
            Vector3D(-sinlambdaint, coslambdaint, 0),
            Vector3D(cosphiint * coslambdaint, cosphiint * sinlambdaint, sinphiint),
        )

        vintsez: Vector3D = rot_rsw_sez.multiply_vector(vintrsw)
        phidotint: float = -vintsez.x / magrint
        lambdadotint: float = vintsez.y / (magrint * cosphiint)

        v_hcw: Vector3D = Vector3D(
            vintsez.z - vtgtrsw.x,
            magrtgt * (lambdadotint - lambdadottgt),
            magrtgt * phidotint,
        )

        return HCW(origin.epoch, r_hcw, v_hcw)


class _StateConvertHCW:
    """class used to perform conversions from the Hill frame"""

    @staticmethod
    def to_gcrf(state: HCW, origin: GCRF) -> GCRF:
        """create an inertial state for the calling state

        :param origin: inertial state that acts as the origin for the relative state
        :type origin: GCRF
        :return: inertial state of the relative spacecraft
        :rtype: GCRF
        """
        magrtgt: float = origin.position.magnitude()
        magrint: float = magrtgt + state.position.x
        rot_eci_rsw: Matrix3D = HCW.frame_matrix(origin)
        vtgtrsw: Vector3D = rot_eci_rsw.multiply_vector(origin.velocity)

        lambdadottgt: float = vtgtrsw.y / magrtgt
        lambdaint: float = state.position.y / magrtgt
        phiint: float = state.position.z / magrtgt
        sinphiint: float = sin(phiint)
        cosphiint: float = cos(phiint)
        sinlambdaint: float = sin(lambdaint)
        coslambdaint: float = cos(lambdaint)

        rot_rsw_sez: Matrix3D = Matrix3D(
            Vector3D(sinphiint * coslambdaint, sinphiint * sinlambdaint, -cosphiint),
            Vector3D(-sinlambdaint, coslambdaint, 0),
            Vector3D(cosphiint * coslambdaint, cosphiint * sinlambdaint, sinphiint),
        )

        rdotint: float = state.velocity.x + vtgtrsw.x
        lambdadotint: float = state.velocity.y / magrtgt + lambdadottgt
        phidotint: float = state.velocity.z / magrtgt
        vintsez: Vector3D = Vector3D(-magrint * phidotint, magrint * lambdadotint * cosphiint, rdotint)
        vintrsw: Vector3D = rot_rsw_sez.transpose().multiply_vector(vintsez)
        vinteci: Vector3D = rot_eci_rsw.transpose().multiply_vector(vintrsw)

        rintrsw: Vector3D = Vector3D(
            cosphiint * magrint * coslambdaint,
            cosphiint * magrint * sinlambdaint,
            sinphiint * magrint,
        )

        rinteci: Vector3D = rot_eci_rsw.transpose().multiply_vector(rintrsw)

        return GCRF(origin.epoch, rinteci, vinteci)


class StateConvert:
    """class used to perform state conversions"""

    #: used to convert from HCW to other frames
    hcw = _StateConvertHCW

    #: used to convert from GCRF to other frames
    gcrf = _StateConvertGCRF



import unittest

from pysmad.math.constants import SECONDS_IN_SIDEREAL_DAY
from pysmad.math.functions import EquationsOfMotion, LegendrePolynomial
from pysmad.math.linalg import Vector3D

MU: float = 398600.4418
GEO_RADIUS: float = 42164.16962995827
GEO_NU: float = 7.292115856392646e-05
A: float = 2
B: float = 1.7320508075688772
E: float = 0.5
P: float = 42139.21213813413
R: float = 42164.16962995827
V: float = 3.07375
PHI: float = 1e-6
H: float = 129602.11640001943


class TestSemiMajorAxis(unittest.TestCase):
    def test_from_mu_n(self):
        self.assertAlmostEqual(EquationsOfMotion.A.from_mu_n(MU, GEO_NU), GEO_RADIUS)

    def test_from_mu_tau(self):
        self.assertAlmostEqual(EquationsOfMotion.A.from_mu_tau(MU, SECONDS_IN_SIDEREAL_DAY), GEO_RADIUS)

    def test_from_mu_r_v(self):
        self.assertAlmostEqual(EquationsOfMotion.A.from_mu_r_v(MU, R, V), 42139.22690208438)


class TestMeanMotion(unittest.TestCase):
    def test_from_a_mu(self):
        self.assertAlmostEqual(EquationsOfMotion.N.from_a_mu(GEO_RADIUS, MU), GEO_NU)

    def test_from_tau(self):
        self.assertAlmostEqual(EquationsOfMotion.N.from_tau(SECONDS_IN_SIDEREAL_DAY), GEO_NU)


class TestSemiMinorAxis(unittest.TestCase):
    def test_from_a_e(self):
        self.assertAlmostEqual(EquationsOfMotion.B.from_a_e(A, E), B)


class TestEccentricity(unittest.TestCase):
    def test_from_a_b(self):
        self.assertAlmostEqual(EquationsOfMotion.E.from_a_b(A, B), E)

    def test_from_a_c(self):
        self.assertAlmostEqual(EquationsOfMotion.E.from_a_c(A, 1), E)

    def test_from_a_p(self):
        self.assertAlmostEqual(EquationsOfMotion.E.from_a_p(GEO_RADIUS, P), 0.024329248415869805)


class TestFlattening(unittest.TestCase):
    def test_from_a_b(self):
        self.assertAlmostEqual(EquationsOfMotion.F.from_a_b(A, B), 0.1339745962155614)


class TestSemiParameter(unittest.TestCase):
    def test_from_mu_h(self):
        self.assertAlmostEqual(EquationsOfMotion.P.from_mu_h(MU, H), P)

    def test_from_a_b(self):
        self.assertAlmostEqual(EquationsOfMotion.P.from_a_b(A, B), 1.5)

    def test_from_a_e(self):
        self.assertAlmostEqual(EquationsOfMotion.P.from_a_e(GEO_RADIUS, 0.024329248415869805), P)


class TestArealVelocity(unittest.TestCase):
    def test_from_mu_p(self):
        self.assertAlmostEqual(EquationsOfMotion.H.from_mu_p(MU, P), H)

    def test_from_r_v_phi(self):
        self.assertAlmostEqual(EquationsOfMotion.H.from_r_v_phi(R, V, PHI), H)

    def test_from_r_v(self):
        h = EquationsOfMotion.H.from_r_v(Vector3D(1, 2, 3), Vector3D(4, 2, 42))
        self.assertAlmostEqual(h.x, 78)
        self.assertAlmostEqual(h.y, -30)
        self.assertAlmostEqual(h.z, -6)


class TestPeriod(unittest.TestCase):
    def test_from_a_mu(self):
        self.assertAlmostEqual(EquationsOfMotion.TAU.from_a_mu(GEO_RADIUS, MU), SECONDS_IN_SIDEREAL_DAY)


class TestVisVivaVelocity(unittest.TestCase):
    def test_from_a_mu_r(self):
        self.assertAlmostEqual(EquationsOfMotion.V.from_a_mu_r(GEO_RADIUS, MU, GEO_RADIUS), 3.07466009930248)

    def test_from_mu_r_e_nu(self):
        self.assertAlmostEqual(EquationsOfMotion.V.from_mu_r_e_nu(MU, GEO_RADIUS, 0, 0), 3.07466009930248)

    def test_from_mu_r_xi(self):
        self.assertAlmostEqual(EquationsOfMotion.V.from_mu_r_xi(MU, R, -4.72956519499274), V)


class TestSpecificMechanicalEnergy(unittest.TestCase):
    def test_from_mu_r_v(self):
        self.assertAlmostEqual(EquationsOfMotion.XI.from_mu_r_v(MU, R, V), -4.72956519499274)

    def test_from_mu_a(self):
        self.assertAlmostEqual(EquationsOfMotion.XI.from_mu_a(MU, A), -99650.11045)


class TestTrueAnomaly(unittest.TestCase):
    def test_from_e_ea(self):
        self.assertAlmostEqual(EquationsOfMotion.NU.from_e_ea(0.025, 0.1), 0.10252766708805071)


class TestEccentricAnomaly(unittest.TestCase):
    def test_from_ma_e(self):
        self.assertAlmostEqual(EquationsOfMotion.EA.from_ma_e(1, 0.2), 1.1853242038613385)

    def test_from_rdv_r_a_n(self):
        self.assertAlmostEqual(EquationsOfMotion.EA.from_rdv_r_a_n(0, GEO_RADIUS, GEO_RADIUS, GEO_NU), 0)


class TestInclination(unittest.TestCase):
    def test_from_w(self):
        self.assertAlmostEqual(EquationsOfMotion.I.from_w(Vector3D(1, 1, 1)), 0.9553166181245093)


class TestMeanAnomaly(unittest.TestCase):
    def test_from_ea_e(self):
        self.assertAlmostEqual(EquationsOfMotion.MA.from_ea_e(1, 0.2), 0.8317058030384207)


class TestArgumentOfPerigee(unittest.TestCase):
    def test_from_u_nu(self):
        self.assertAlmostEqual(EquationsOfMotion.W.from_u_nu(1, 2), 5.283185307179586)


class TestArgumentOfLatitude(unittest.TestCase):
    def test_from_r_w(self):
        self.assertAlmostEqual(EquationsOfMotion.U.from_r_w(Vector3D(1, 2, 3), Vector3D(4, 2, 42)), 0.4636476090008061)


class TestRAAN(unittest.TestCase):
    def test_from_w(self):
        self.assertAlmostEqual(EquationsOfMotion.RAAN.from_w(Vector3D(1, 1, 1)), 2.356194490192345)


class TestLegendrePolynomial(unittest.TestCase):
    def test_p(self):
        p = LegendrePolynomial(1).p
        self.assertAlmostEqual(p[0][0], 1)
        self.assertAlmostEqual(p[1][0], 0.8414709848078965)
        self.assertAlmostEqual(p[1][1], 0.5403023058681398)
        self.assertAlmostEqual(p[2][0], 0.5621101274103568)
        self.assertAlmostEqual(p[2][1], 1.3639461402385225)
        self.assertAlmostEqual(p[2][2], 0.8757797451792866)
        self.assertAlmostEqual(p[3][0], 0.2273516142655441)
        self.assertAlmostEqual(p[3][1], 2.0588492958263878)
        self.assertAlmostEqual(p[3][2], 3.6847162232541146)
        self.assertAlmostEqual(p[3][3], 2.365929078764902)
        self.assertAlmostEqual(p[4][0], -0.08679046873880569)
        self.assertAlmostEqual(p[4][1], 2.2238163502521444)
        self.assertAlmostEqual(p[4][2], 8.66258689896924)
        self.assertAlmostEqual(p[4][3], 13.936024703257592)
        self.assertAlmostEqual(p[4][4], 8.948218557440123)



class Alerts:

    WARNING: str = "\x1b[1;33;40m"
    CLOSE: str = "\x1b[0m"
    PROCESS: str = "\x1b[1;36;40m"
    SUCCESS: str = "\x1b[1;32;40m"

    @staticmethod
    def color(color_code: str, msg) -> None:
        print("".join([color_code, msg, Alerts.CLOSE]))

    @staticmethod
    def print_query_max_warning() -> None:
        msg: str = "WARNING:  Results may be missing.  The returned volume is equal to the query maximum."
        Alerts.color(Alerts.WARNING, msg)

    @staticmethod
    def print_query_start(url: str) -> None:
        Alerts.color(Alerts.PROCESS, " ".join(["Query in progress for URL", url]))

    @staticmethod
    def print_successful_query(num: int) -> None:
        msg: str = " ".join(["SUCCESS:  Query returned", str(num), "results."])
        Alerts.color(Alerts.SUCCESS, msg)



import matplotlib.pyplot as plt

from pysmad.bodies.artificial import Spacecraft
from pysmad.coordinates.states import GCRF, HCW, StateConvert
from pysmad.math.linalg import Vector3D, Vector6D
from pysmad.time import Epoch

# Define the initial time reference
start_epoch: Epoch = Epoch.from_gregorian(2022, 12, 20, 0, 0, 0)

# Create a desired relative state for the chase vehicle
rel_chase_state = HCW.from_state_vector(Vector6D(-11, 0, 0, 0, 0, 0))

# Create an inertial state for the target vehicle
target_state: GCRF = GCRF(start_epoch, Vector3D(42164, 0, 0), Vector3D(0, 3.075, 0))

# Create a spacecraft to act as an estimated state of the target spacecraft which will be used to initialize the filter
seed: Spacecraft = Spacecraft(GCRF(start_epoch, Vector3D(42164.5, 0.5, 0.5), Vector3D(0, 3.075, 0)))

# Create the chase vehicle's inertial state using the target state as the origin
chase_state: GCRF = StateConvert.hcw.to_gcrf(rel_chase_state, target_state)

# Define a propagation epoch of one day
end_epoch = start_epoch.plus_days(1)

# Create the target and chase spacecraft
chase: Spacecraft = Spacecraft(chase_state)
target: Spacecraft = Spacecraft(target_state)

# Create empty lists to store data
times, err = [], []

# Initialize the kalman filter with the estimated state
chase.acquire(seed)

# Propagate
while chase.current_epoch().value < end_epoch.value:

    # Increment the epochs of the spacecraft
    chase.step()
    target.step_to_epoch(chase.current_epoch())

    # Update chase attitude to point payloads at target
    chase.track_state(target)

    # Create a simulated ob and feed it to the filter
    chase.process_wfov(target)

    # Get the relative truth state
    ric = chase.hill_position(target)

    # Get the relative estimate from the chase solution
    nav = chase.filter.propagator.state.position

    # Store data
    times.append(chase.current_epoch().value)
    err.append(abs(nav.magnitude() - ric.magnitude()))

# Plot results
plt.plot(times, err)
plt.show()



from math import ceil, e, log
from typing import List

from pysmad.coordinates.states import GCRF
from pysmad.math.constants import SEA_LEVEL_G, SECONDS_IN_DAY
from pysmad.math.linalg import Vector3D
from pysmad.time import Epoch


class RK4:

    #: Largest step to be taken by the integrator
    MAX_STEP = 300

    def __init__(self, state: GCRF) -> None:
        """class used to propagate a satellite state

        :param state: ECI state of the satellite to be propagated
        :type state: GCRF
        """
        #: the current state of the propagator
        self.state: GCRF = state.copy()

        #: integration step to be taken when the propagator is advanced
        self.step_size: float = RK4.MAX_STEP

        #: mass flow rate used to apply thrusts to propagator
        self.m_dot: float = 0

        #: initial mass when using the propagator to apply thrust
        self.m0: float = 0

        #: gcrf direction of any applied thrusts
        self.thrust_direction: Vector3D = Vector3D(0, 0, 0)

        #: specific impulse used to apply thrusts
        self.isp: float = 0

    def step(self) -> None:
        """advance the propagator state by the stored time step"""
        h = self.step_size

        epoch_0: Epoch = self.state.epoch.copy()

        self.state.thrust = self.thrust_vector(0)
        y: List[Vector3D] = self.state.vector_list()

        k1: List[Vector3D] = self.state.derivative()

        dsecs: float = h / 2
        ddays: float = dsecs / SECONDS_IN_DAY
        epoch_1 = epoch_0.plus_days(ddays)
        y1: GCRF = GCRF(epoch_1, y[0].plus(k1[0].scaled(dsecs)), y[1].plus(k1[1].scaled(dsecs)))
        y1.thrust = self.thrust_vector(dsecs)
        k2: List[Vector3D] = y1.derivative()

        y2: GCRF = GCRF(epoch_1, y[0].plus(k2[0].scaled(dsecs)), y[1].plus(k2[1].scaled(dsecs)))
        y2.thrust = self.thrust_vector(dsecs)
        k3: List[Vector3D] = y2.derivative()

        epoch_2 = epoch_1.plus_days(ddays)
        y3: GCRF = GCRF(epoch_2, y[0].plus(k3[0].scaled(h)), y[1].plus(k3[1].scaled(h)))
        y3.thrust = self.thrust_vector(dsecs * 2)
        k4: List[Vector3D] = y3.derivative()

        coeff: float = 1 / 6
        dv: Vector3D = k1[0].plus(k2[0].scaled(2).plus(k3[0].scaled(2).plus(k4[0]))).scaled(coeff)
        da: Vector3D = k1[1].plus(k2[1].scaled(2).plus(k3[1].scaled(2).plus(k4[1]))).scaled(coeff)

        self.state = GCRF(
            epoch_2,
            self.state.position.plus(dv.scaled(h)),
            self.state.velocity.plus(da.scaled(h)),
        )

    def maneuver(self, gcrf_thrust: Vector3D, m_dot: float, m0: float, isp: float) -> None:
        """propagate the state using continuous thrust principles

        :param gcrf_thrust: components of the maneuver in the gcrf frame
        :type gcrf_thrust: Vector3D
        :param m_dot: mass flow rate
        :type m_dot: float
        :param m0: initial mass
        :type m0: float
        :param isp: specific impulse
        :type isp: float
        """
        m_spec: float = m_dot / m0
        dv_duration: float = (1 / m_spec) * (
            1 - e ** (m_spec * gcrf_thrust.magnitude() / (-isp * m_spec * SEA_LEVEL_G))
        )
        self.thrust_direction = gcrf_thrust.copy()
        self.m0 = m0
        self.m_dot = m_dot
        self.isp = isp
        self.step_to_epoch(self.state.epoch.plus_days(dv_duration / SECONDS_IN_DAY))
        self.m0 = 0
        self.m_dot = 0

    def step_to_epoch(self, epoch: Epoch) -> None:
        """advance the propagator state to the argument epoch

        :param epoch: time of state to be calculated
        :type epoch: Epoch
        """

        # Calculate time delta in seconds
        dt = (epoch.value - self.state.epoch.value) * SECONDS_IN_DAY

        # Determine number of steps required to meet new epoch while staying below the maximum step
        num_steps = ceil(abs(dt / self.MAX_STEP))

        # Store current step size
        old_step = self.step_size

        # Temporarily set step size to calculated dt
        if num_steps > 0:
            self.step_size = dt / num_steps

        # Step until desired epoch is achieved
        step_n = 0
        while step_n < num_steps:
            self.step()
            step_n += 1

        # Reset step size
        self.step_size = old_step

    def thrust_vector(self, dt: float) -> Vector3D:
        """calculate the acceleration vector due to thrust

        :param dt: time step at which to calculate the thrust
        :type dt: float
        :return: thrust vector with gcrf components
        :rtype: Vector3D
        """
        a: Vector3D = Vector3D(0, 0, 0)
        if self.m_dot != 0:
            if dt == 0:
                a = self.thrust_direction.normalized().scaled(self.m_dot * self.isp * SEA_LEVEL_G / self.m0)
            else:
                ln = log(1 - self.m_dot * dt / self.m0)
                mt = self.m0 - self.m_dot * dt
                f = self.m_dot * self.isp * SEA_LEVEL_G
                dv: Vector3D = self.thrust_direction.normalized().scaled((-f / self.m_dot) * ln)
                a = dv.scaled((self.m_dot / mt) * (1 / (-log(1 - self.m_dot * dt / self.m0))))

        return a



from math import asin, atan2, cos, pi, sin, sqrt

from pysmad.bodies.celestial import Earth
from pysmad.coordinates.elements import ClassicalElements
from pysmad.estimation.obs import Observation
from pysmad.math.constants import SECONDS_IN_DAY
from pysmad.math.linalg import Vector3D


class Gauss:

    NU_TOLERANCE: float = 1e-12

    @staticmethod
    def coes_from_positions(ob1: Observation, ob2: Observation) -> ClassicalElements:
        """calculate the orbital elements of a satellite using the gauss method for two position vectors

        :param ob1: first observation
        :type ob1: Observation
        :param ob2: second observation
        :type ob2: Observation
        :return: orbital elements for a spacecraft that travels through the two positions
        :rtype: ClassicalElements
        """

        # method is found in Chapter 2 of Satellite Orbits
        ra: Vector3D = ob1.ijk_position()
        rb: Vector3D = ob2.ijk_position()
        ra_mag: float = ra.magnitude()
        rb_mag: float = rb.magnitude()

        # equation 2.109
        ea: Vector3D = ra.normalized()

        # equation 2.110
        r0: Vector3D = rb.minus(ea.scaled(rb.dot(ea)))
        r0_mag: float = r0.magnitude()
        e0: Vector3D = r0.normalized()

        # equation 2.111
        w: Vector3D = ea.cross(e0)
        ua: float = atan2(ra.z, -ra.x * w.y + ra.y * w.x)
        inc: float = atan2(sqrt(w.x * w.x + w.y * w.y), w.z)
        raan: float = atan2(w.x, -w.y)
        if inc < 0:
            inc += 2 * pi
        if raan < 0:
            raan += 2 * pi

        # equation 2.99
        tau: float = sqrt(Earth.MU) * (ob2.epoch().value - ob1.epoch().value) * SECONDS_IN_DAY

        # equation 2.101
        base: float = 2 * (ra_mag * rb_mag + ra.dot(rb))
        m: float = tau * tau * (base) ** -1.5
        l: float = (ra_mag + rb_mag) / (2 * sqrt(base)) - 0.5

        # equation 2.107
        n0: float = 12 / 22 + 10 / 22 * sqrt(1 + (44 / 9) * (m / (l + 5 / 6)))
        n1: float = n0 + 0.1
        n2: float = n0
        n: float = Gauss.solve_nu(m, l, n1, n2)

        # equation 2.112
        base = ra_mag * r0_mag * n / tau
        p: float = base * base

        # equation 2.114
        ecosva: float = p / ra_mag - 1
        ecosvb: float = p / rb_mag - 1

        # equation 2.116
        esinva: float = (ecosva * (rb.dot(ea) / rb_mag) - ecosvb) / (r0_mag / rb_mag)

        ta: float = atan2(esinva, ecosva)
        if ta < 0:
            ta += 2 * pi

        e: float = ecosva / cos(ta)

        aop: float = ua - ta
        if aop < 0:
            aop += 2 * pi

        sma: float = p / (1 - e * e)

        e_anom: float = atan2(sqrt(1 - e * e) * sin(ta), cos(ta) + e)
        ma: float = e_anom - e * sin(e_anom)

        return ClassicalElements(ob1.epoch(), sma, e, inc, raan, aop, ma)

    @staticmethod
    def solve_nu(m: float, l: float, n0: float, n1: float) -> float:
        """recursively solve for nu in equation 2.105 of Satellite Orbits (secant procedure)

        :param m: auxiliary variable
        :type m: float
        :param l: auxiliary variable
        :type l: float
        :param n0: nu at index i - 1
        :type n0: float
        :param n1: nu at index i
        :type n1: float
        :return: nu at index i + 1
        :rtype: float
        """
        converged: bool = False
        f_of_nu_1: float
        n2: float
        while not converged:
            f_of_nu_1 = Gauss.f_of_nu(n1, m, l)
            n2 = n1 - f_of_nu_1 * (n1 - n0) / (f_of_nu_1 - Gauss.f_of_nu(n0, m, l))
            if abs(n2 - n1) < Gauss.NU_TOLERANCE:
                converged = True
            else:
                n0 = n1
                n1 = n2

        return n2

    @staticmethod
    def f_of_nu(n: float, m: float, l: float) -> float:
        """equation 2.106 in Satellite orbits

        :param n: nu
        :type n: float
        :param m: auxiliary variable
        :type m: float
        :param l: auxiliary variable
        :type l: float
        :return: f(nu)
        :rtype: float
        """
        m_over_nu_squared: float = m / (n * n)
        return 1 - n + m_over_nu_squared * Gauss.w_of_omega(m_over_nu_squared - l)

    @staticmethod
    def w_of_omega(w: float) -> float:
        """equation 2.103 in Satellite Orbits

        :param w: omega
        :type w: float
        :return: w(omega)
        :rtype: float
        """
        g: float = 2 * asin(sqrt(w))
        sg: float = sin(g)
        return (2 * g - sin(2 * g)) / (sg * sg * sg)



import unittest

from pysmad.planning.maneuvers import Hohmann, VisViva


class TestVisViva(unittest.TestCase):
    def test_circular_velocity(self):
        self.assertAlmostEqual(VisViva.circular_velocity(42164), 3.074666284127684)

    def test_eccentric_velocity(self):
        self.assertAlmostEqual(VisViva.eccentric_velocity(42164, 42244), 3.077576247190643)


class TestHohmann(unittest.TestCase):
    PLANNER: Hohmann = Hohmann(42164, 42324)

    def test_planner(self):
        self.assertAlmostEqual(self.PLANNER.delta_v_1, 0.0029099630629589868)
        self.assertAlmostEqual(self.PLANNER.delta_v_2, 0.0029072089766200016)
